import unittest
import numpy as np
import functools

from rlcard.games.doudizhu.game import DoudizhuGame as Game
from rlcard.games.doudizhu.utils import get_landlord_score, encode_cards
from rlcard.games.doudizhu.utils import doudizhu_sort_str
from rlcard.games.doudizhu.judger import DoudizhuJudger as Judger


class TestDoudizhuGame(unittest.TestCase):

    def test_get_num_players(self):
        game = Game()
        num_players = game.get_num_players()
        self.assertEqual(num_players, 3)

    def test_get_num_actions(self):
        game = Game()
        num_actions = game.get_num_actions()
        self.assertEqual(num_actions, 27472)

    def test_init_game(self):
        game = Game()
        state, current_player = game.init_game()
        total_cards = list(state['current_hand'] + state['others_hand'])
        total_cards.sort(key=functools.cmp_to_key(doudizhu_sort_str))
        deck = list(game.round.deck_str)
        self.assertEqual(state['self'], current_player)
        self.assertIs(len(''.join(state['played_cards'])), 0)
        self.assertEqual(len(total_cards), 54)
        self.assertListEqual(total_cards, deck)

    def test_step(self):
        game = Game()
        state, _ = game.init_game()
        action = state['actions'][0]
        state, next_player_id = game.step(action)
        next_player = game.players[next_player_id]
        player_id = (next_player.player_id-1)%len(game.players)
        self.assertEqual(state['trace'][0][0], player_id)
        self.assertEqual(state['trace'][0][1], action)

    def test_get_player_id(self):
        game = Game()
        _, player_id = game.init_game()
        current_player_id = game.get_player_id()
        self.assertEqual(current_player_id, player_id)

    def test_proceed_game(self):
        game = Game()
        state, player_id = game.init_game()
        while not game.is_over():
            action = np.random.choice(list(state['actions']))
            state, next_player_id = game.step(action)
            player = game.players[player_id]
            self.assertEqual((player.player_id+1)%len(game.players), next_player_id)
            player_id = next_player_id
        for player_id in range(3):
            state = game.get_state(player_id)
            self.assertEqual(state['actions'], [])

    def test_step_back(self):
        #case 1: action, stepback
        game = Game(allow_step_back=True)
        state, player_id = game.init_game()
        action = state['actions'][0]
        playable_cards = game.judger.playable_cards
        game.step(action)
        game.step_back()
        self.assertEqual(game.judger.playable_cards, playable_cards)
        self.assertEqual(game.round.greater_player, None)
        self.assertEqual(game.round.current_player, player_id)
        self.assertEqual(len(game.history), 0)
        game.state['actions'].sort()
        state['actions'].sort()
        #self.assertEqual(game.state, state)
        self.assertEqual(game.step_back(), False)

        #case 2: action, pass, stepback
        game = Game(allow_step_back=True)
        state, player_id = game.init_game()
        action = state['actions'][0]
        game.step(action)
        actions = game.state['actions']
        playable_cards = game.judger.playable_cards
        played_cards = game.players[game.round.current_player].played_cards
        game.step('pass')
        game.step_back()
        #judger.playable_cards should be the same
        self.assertEqual(game.judger.playable_cards, playable_cards)
        #players[current_player].played_cards should be the same
        self.assertEqual(game.players[game.round.current_player].played_cards, played_cards)
        #greater_player should be the same
        self.assertEqual(game.round.greater_player.player_id, 0)
        actions.sort()
        game.state['actions'].sort()
        #actions should be the same after step_back()
        self.assertEqual(game.state['actions'], actions)

        #case 3: action, pass, pass, action, stepback
        game = Game(allow_step_back=True)
        state, player_id = game.init_game()
        action = state['actions'][0]
        game.step(action)
        game.step('pass')
        game.step('pass')
        actions = game.state['actions']
        playable_cards = game.judger.playable_cards
        played_cards = game.players[game.round.current_player].played_cards
        game.step(actions[0])
        game.step_back()
        #judger.playable_cards should be the same
        self.assertEqual(game.judger.playable_cards, playable_cards)
        #players[current_player].played_cards should be the same
        self.assertEqual(game.players[game.round.current_player].played_cards, played_cards)
        #greater_player should be the same
        self.assertEqual(game.round.greater_player.player_id, 0)
        actions.sort()
        game.state['actions'].sort()
        #actions should be the same after step_back()
        self.assertEqual(game.state['actions'], actions)
        game.step_back()
        #greater_player should be the same
        self.assertEqual(game.round.greater_player.player_id, 0)

    def test_get_landlord_score(self):
        score_1 = get_landlord_score('56888TTQKKKAA222R')
        self.assertEqual(score_1, 12)

    def test_encode_cards(self):
        plane = np.zeros((5, 15), dtype=int)
        plane[0] = np.ones(15, dtype=int)
        cards = '333BR'
        encode_cards(plane, cards)
        self.assertEqual(plane[3][0], 1)
        self.assertEqual(plane[1][13], 1)
        self.assertEqual(plane[1][14], 1)

    def test_judge_payoffs(self):
        payoffs = Judger.judge_payoffs(0, 0)
        self.assertEqual(payoffs[0], 1)
        payoffs = Judger.judge_payoffs(2, 0)
        self.assertEqual(payoffs[0], 1)
        self.assertEqual(payoffs[1], 1)

if __name__ == '__main__':
    unittest.main()



'''
    File name: bridge/judger.py
    Author: William Hale
    Date created: 11/25/2021
'''

from typing import List

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game import BridgeGame

from .utils.action_event import PlayCardAction
from .utils.action_event import ActionEvent, BidAction, PassAction, DblAction, RdblAction
from .utils.move import MakeBidMove, MakeDblMove, MakeRdblMove
from .utils.bridge_card import BridgeCard


class BridgeJudger:

    '''
        Judger decides legal actions for current player
    '''

    def __init__(self, game: 'BridgeGame'):
        ''' Initialize the class BridgeJudger
        :param game: BridgeGame
        '''
        self.game: BridgeGame = game

    def get_legal_actions(self) -> List[ActionEvent]:
        """
        :return: List[ActionEvent] of legal actions
        """
        legal_actions: List[ActionEvent] = []
        if not self.game.is_over():
            current_player = self.game.round.get_current_player()
            if not self.game.round.is_bidding_over():
                legal_actions.append(PassAction())
                last_make_bid_move: MakeBidMove or None = None
                last_dbl_move: MakeDblMove or None = None
                last_rdbl_move: MakeRdblMove or None = None
                for move in reversed(self.game.round.move_sheet):
                    if isinstance(move, MakeBidMove):
                        last_make_bid_move = move
                        break
                    elif isinstance(move, MakeRdblMove):
                        last_rdbl_move = move
                    elif isinstance(move, MakeDblMove) and not last_rdbl_move:
                        last_dbl_move = move
                first_bid_action_id = ActionEvent.first_bid_action_id
                next_bid_action_id = last_make_bid_move.action.action_id + 1 if last_make_bid_move else first_bid_action_id
                for bid_action_id in range(next_bid_action_id, first_bid_action_id + 35):
                    action = BidAction.from_action_id(action_id=bid_action_id)
                    legal_actions.append(action)
                if last_make_bid_move and last_make_bid_move.player.player_id % 2 != current_player.player_id % 2 and not last_dbl_move and not last_rdbl_move:
                    legal_actions.append(DblAction())
                if last_dbl_move and last_dbl_move.player.player_id % 2 != current_player.player_id % 2:
                    legal_actions.append(RdblAction())
            else:
                trick_moves = self.game.round.get_trick_moves()
                hand = self.game.round.players[current_player.player_id].hand
                legal_cards = hand
                if trick_moves and len(trick_moves) < 4:
                    led_card: BridgeCard = trick_moves[0].card
                    cards_of_led_suit = [card for card in hand if card.suit == led_card.suit]
                    if cards_of_led_suit:
                        legal_cards = cards_of_led_suit
                for card in legal_cards:
                    action = PlayCardAction(card=card)
                    legal_actions.append(action)
        return legal_actions



'''
    Project: Gui Gin Rummy
    File name: utils.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from typing import List

import tkinter as tk

import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils

from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError

from .canvas_item import CardItem, CanvasItem
from .player_type import PlayerType

from .configurations import SCORE_PLAYER_0_ACTION_ID, SCORE_PLAYER_1_ACTION_ID
from .configurations import DRAW_CARD_ACTION_ID, PICK_UP_DISCARD_ACTION_ID
from .configurations import DECLARE_DEAD_HAND_ACTION_ID
from .configurations import DISCARD_ACTION_ID, KNOCK_ACTION_ID

from . import configurations


def is_debug() -> bool:
    result = __debug__ and configurations.IS_DEBUG
    return result


def gin_rummy_sort_order_id(card_id: int) -> int:
    return 4 * (card_id % 13) + (3 - card_id // 13)  # would have been better if card_id was specified in order wanted


def move_to(item_id: int, x: int, y: int, parent: tk.Canvas):
    parent.coords(item_id, x, y)


def translated_by(dx: float, dy: float, location):
    if not len(location) == 2:
        raise GinRummyProgramError("location={} must have length of 2.".format(location))
    return [location[0] + dx, location[1] + dy]


def player_name(player_id: int) -> str:
    return "North" if player_id == 0 else "South" if player_id == 1 else "X"


def player_short_name(player_id: int) -> str:
    return "N" if player_id == 0 else "S" if player_id == 1 else "X"


def get_action_type(action: int) -> int:
    if action == DRAW_CARD_ACTION_ID:
        result = action
    elif action == PICK_UP_DISCARD_ACTION_ID:
        result = action
    elif action == DECLARE_DEAD_HAND_ACTION_ID:
        result = action
    elif DISCARD_ACTION_ID <= action < DISCARD_ACTION_ID + 52:
        result = DISCARD_ACTION_ID
    elif KNOCK_ACTION_ID <= action < KNOCK_ACTION_ID + 52:
        result = KNOCK_ACTION_ID
    elif action == SCORE_PLAYER_0_ACTION_ID:
        result = action
    elif action == SCORE_PLAYER_1_ACTION_ID:
        result = action
    else:
        raise GinRummyProgramError("No action type for {}.".format(action))
    return result


def get_action_card_id(action: int) -> int or None:
    result = None
    action_type = get_action_type(action)
    if action_type == DISCARD_ACTION_ID:
        result = action - DISCARD_ACTION_ID
    elif action_type == KNOCK_ACTION_ID:
        result = action - KNOCK_ACTION_ID
    return result


#   =========================================
#   Transformations (i.e. complex setter)
#   =========================================

def set_card_id_face_up(card_id: int, face_up: bool, game_canvas: 'GameCanvas'):
    card = gin_rummy_utils.card_from_card_id(card_id=card_id)
    card_image = game_canvas.card_images[card.rank, card.suit]
    if card_image.face_up != face_up:
        card_item_id = game_canvas.card_item_ids[card_id]
        target_image = card_image if face_up else game_canvas.card_back_image
        game_canvas.itemconfig(card_item_id, image=target_image)
        card_image.face_up = face_up


def flip_card_id(card_id: int, game_canvas: 'GameCanvas'):
    card = gin_rummy_utils.card_from_card_id(card_id=card_id)
    card_image = game_canvas.card_images[card.rank, card.suit]
    card_item_id = game_canvas.card_item_ids[card_id]
    card_image.face_up = not card_image.face_up
    target_image = card_image if card_image.face_up else game_canvas.card_back_image
    game_canvas.itemconfig(card_item_id, image=target_image)


def jog_card_id(card_id: int, dx: float, dy: float, game_canvas: 'GameCanvas'):
    # jog card if held by human player (allows computer to not reveal location of selected cards in hand)
    # TODO: configuration option to highlight card rather than jogging it
    card_item_id = game_canvas.card_item_ids[card_id]
    card_item_id_tags = game_canvas.getter.get_tags(item_id=card_item_id)
    player_id = 0 if configurations.NORTH_HELD_PILE_TAG in card_item_id_tags else 1
    if game_canvas.player_types[player_id] is PlayerType.human_player:
        game_canvas.move(card_item_id, dx, dy)


def drop_item_ids(item_ids: List[int], on_item_id: int, player_id: int, game_canvas: 'GameCanvas'):
    # on_item_id must be in held_pile_item_ids of player_id
    # item_ids are inserted into held_pile of player_id after on_item_id
    held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id)
    held_pile_ghost_card_item = game_canvas.held_pile_ghost_card_items[player_id]
    if not (on_item_id == held_pile_ghost_card_item or on_item_id in held_pile_item_ids):
        raise GinRummyProgramError("on_item_id={} is invalid drop location.".format(on_item_id))
    held_pile_item_ids_count = len(held_pile_item_ids)
    held_pile_tag = game_canvas.held_pile_tags[player_id]
    on_item_index = -1 if on_item_id == held_pile_ghost_card_item else held_pile_item_ids.index(on_item_id)
    after_item_ids = []
    for i in range(on_item_index + 1, held_pile_item_ids_count):
        after_item_id = held_pile_item_ids[i]
        if after_item_id not in item_ids:
            after_item_ids.append(after_item_id)
    drop_canvas_items = [game_canvas.canvas_item_by_item_id[item_id] for item_id in item_ids]
    drop_card_items = [x for x in drop_canvas_items if isinstance(x, CardItem)]
    sorted_drop_card_items = sorted(drop_card_items, reverse=True,
                                    key=lambda card_item: gin_rummy_sort_order_id(card_item.card_id))
    sorted_item_ids = [x.item_id for x in sorted_drop_card_items]
    for item_id in sorted_item_ids:
        item_tags = game_canvas.getter.get_tags(item_id)
        if configurations.SELECTED_TAG in item_tags:
            game_canvas.dtag(item_id, configurations.SELECTED_TAG)
            game_canvas.dtag(item_id, configurations.JOGGED_TAG)
        if configurations.DRAWN_TAG in item_tags:
            game_canvas.dtag(item_id, configurations.DRAWN_TAG)
            if configurations.DISCARD_PILE_TAG in item_tags:
                game_canvas.dtag(item_id, configurations.DISCARD_PILE_TAG)
            elif configurations.STOCK_PILE_TAG in item_tags:
                game_canvas.dtag(item_id, configurations.STOCK_PILE_TAG)
            if held_pile_tag in item_tags:
                raise GinRummyProgramError("item_tags should not contain held_pile_tag.")
            game_canvas.addtag_withtag(held_pile_tag, item_id)
        game_canvas.tag_raise(item_id)
    for after_item_id in after_item_ids:
        game_canvas.tag_raise(after_item_id)
    fan_held_pile(player_id, game_canvas=game_canvas)


def fan_held_pile(player_id: int, game_canvas: 'GameCanvas'):
    held_pile_tab = game_canvas.held_pile_tab
    held_pile_anchor = game_canvas.player_held_pile_anchors[player_id]
    held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=player_id)
    # right justify hand when melding
    dx = 0
    count = len(held_pile_item_ids)
    if count < 10:
        dx = (10 - count) * game_canvas.held_pile_tab
    x = held_pile_anchor[0] + dx
    y = held_pile_anchor[1]
    for held_pile_item_id in held_pile_item_ids:
        move_to(held_pile_item_id, x, y, parent=game_canvas)
        game_canvas.tag_raise(held_pile_item_id)
        x += held_pile_tab


def held_pile_insert(card_item_id: int, above_hit_item_id: int or None, player_id: int, game_canvas: 'GameCanvas'):
    held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=player_id)
    held_pile_item_ids_count = len(held_pile_item_ids)
    if above_hit_item_id is None or above_hit_item_id == game_canvas.held_pile_ghost_card_items[player_id]:
        insertion_index = 0
    else:
        insertion_index = held_pile_item_ids.index(above_hit_item_id) + 1
    held_pile_tab = game_canvas.held_pile_tab
    if not card_item_id == held_pile_item_ids[-1]:  # Note: card_item_id is last and already positioned and raised
        raise GinRummyProgramError("card_item_id={} must be last card of hand.".format(card_item_id))
    for i in range(insertion_index, held_pile_item_ids_count - 1):
        held_pile_item_id = held_pile_item_ids[i]
        game_canvas.move(held_pile_item_id, held_pile_tab, 0)
        game_canvas.tag_raise(held_pile_item_id)


def set_card_item_id_face_up(card_item_id: int, face_up: bool, game_canvas: 'GameCanvas'):
    card_id = game_canvas.card_item_ids.index(card_item_id)
    set_card_id_face_up(card_id=card_id, face_up=face_up, game_canvas=game_canvas)


def toggle_discard_pile_item_selected(game_canvas: 'GameCanvas'):
    current_player_id = game_canvas.current_player_id
    top_discard_pile_item_id = game_canvas.getter.get_top_discard_pile_item_id()
    item_tags = game_canvas.getter.get_tags(top_discard_pile_item_id)
    is_drawn = configurations.DRAWN_TAG in item_tags
    card_id = game_canvas.card_item_ids.index(top_discard_pile_item_id)
    dx = -20 * game_canvas.scale_factor
    dy = 20 * game_canvas.scale_factor
    if current_player_id == 0:
        dy = -dy
    if is_drawn:
        dx = -dx
        dy = -dy
    jog_card_id(card_id=card_id, dx=dx, dy=dy, game_canvas=game_canvas)
    if is_drawn:
        game_canvas.dtag(top_discard_pile_item_id, configurations.DRAWN_TAG)
    else:
        game_canvas.addtag_withtag(configurations.DRAWN_TAG, top_discard_pile_item_id)


def toggle_held_pile_item_selected(item: CanvasItem, game_canvas: 'GameCanvas'):
    if isinstance(item, CardItem):  # don't mess with ghost card
        item_tags = item.get_tags()
        is_selected = configurations.SELECTED_TAG in item_tags
        card_id = item.card_id
        dx = 20 * game_canvas.scale_factor
        dy = -20 * game_canvas.scale_factor
        if is_selected:
            dx = -dx
            dy = -dy
        jog_card_id(card_id=card_id, dx=dx, dy=dy, game_canvas=game_canvas)
        if is_selected:
            game_canvas.dtag(item.item_id, configurations.SELECTED_TAG)
            game_canvas.dtag(item.item_id, configurations.JOGGED_TAG)
        else:
            game_canvas.addtag_withtag(configurations.SELECTED_TAG, item.item_id)
            game_canvas.addtag_withtag(configurations.JOGGED_TAG, item.item_id)


def toggle_stock_pile_item_selected(game_canvas: 'GameCanvas'):
    current_player_id = game_canvas.current_player_id
    top_stock_pile_item_id = game_canvas.getter.get_top_stock_pile_item_id()
    item_tags = game_canvas.getter.get_tags(top_stock_pile_item_id)
    is_drawn = configurations.DRAWN_TAG in item_tags
    card_id = game_canvas.card_item_ids.index(top_stock_pile_item_id)
    flip_card_id(card_id, game_canvas=game_canvas)
    dx = 20 * game_canvas.scale_factor
    dy = 20 * game_canvas.scale_factor
    if current_player_id == 0:
        dy = -dy
    if is_drawn:
        dx = -dx
        dy = -dy
    jog_card_id(card_id=card_id, dx=dx, dy=dy, game_canvas=game_canvas)
    if is_drawn:
        game_canvas.dtag(top_stock_pile_item_id, configurations.DRAWN_TAG)
    else:
        game_canvas.addtag_withtag(configurations.DRAWN_TAG, top_stock_pile_item_id)



# -*- coding: utf-8 -*-
''' Implement Doudizhu Game class
'''
import functools
from heapq import merge
import numpy as np

from rlcard.games.doudizhu.utils import cards2str, doudizhu_sort_card, CARD_RANK_STR
from rlcard.games.doudizhu import Player
from rlcard.games.doudizhu import Round
from rlcard.games.doudizhu import Judger


class DoudizhuGame:
    ''' Provide game APIs for env to run doudizhu and get corresponding state
    information.
    '''
    def __init__(self, allow_step_back=False):
        self.allow_step_back = allow_step_back
        self.np_random = np.random.RandomState()
        self.num_players = 3

    def init_game(self):
        ''' Initialize players and state.

        Returns:
            dict: first state in one game
            int: current player's id
        '''
        # initialize public variables
        self.winner_id = None
        self.history = []

        # initialize players
        self.players = [Player(num, self.np_random)
                        for num in range(self.num_players)]

        # initialize round to deal cards and determine landlord
        self.played_cards = [np.zeros((len(CARD_RANK_STR), ), dtype=np.int32)
                                for _ in range(self.num_players)]
        self.round = Round(self.np_random, self.played_cards)
        self.round.initiate(self.players)

        # initialize judger
        self.judger = Judger(self.players, self.np_random)

        # get state of first player
        player_id = self.round.current_player
        self.state = self.get_state(player_id)

        return self.state, player_id

    def step(self, action):
        ''' Perform one draw of the game

        Args:
            action (str): specific action of doudizhu. Eg: '33344'

        Returns:
            dict: next player's state
            int: next player's id
        '''
        if self.allow_step_back:
            # TODO: don't record game.round, game.players, game.judger if allow_step_back not set
            pass

        # perfrom action
        player = self.players[self.round.current_player]
        self.round.proceed_round(player, action)
        if (action != 'pass'):
            self.judger.calc_playable_cards(player)
        if self.judger.judge_game(self.players, self.round.current_player):
            self.winner_id = self.round.current_player
        next_id = (player.player_id+1) % len(self.players)
        self.round.current_player = next_id

        # get next state
        state = self.get_state(next_id)
        self.state = state

        return state, next_id

    def step_back(self):
        ''' Return to the previous state of the game

        Returns:
            (bool): True if the game steps back successfully
        '''
        if not self.round.trace:
            return False

        #winner_id will be always None no matter step_back from any case
        self.winner_id = None

        #reverse round
        player_id, cards = self.round.step_back(self.players)

        #reverse player
        if (cards != 'pass'):
            self.players[player_id].played_cards = self.round.find_last_played_cards_in_trace(player_id)
        self.players[player_id].play_back()

        #reverse judger.played_cards if needed
        if (cards != 'pass'):
            self.judger.restore_playable_cards(player_id)

        self.state = self.get_state(self.round.current_player)
        return True

    def get_state(self, player_id):
        ''' Return player's state

        Args:
            player_id (int): player id

        Returns:
            (dict): The state of the player
        '''
        player = self.players[player_id]
        others_hands = self._get_others_current_hand(player)
        num_cards_left = [len(self.players[i].current_hand) for i in range(self.num_players)]
        if self.is_over():
            actions = []
        else:
            actions = list(player.available_actions(self.round.greater_player, self.judger))
        state = player.get_state(self.round.public, others_hands, num_cards_left, actions)

        return state

    @staticmethod
    def get_num_actions():
        ''' Return the total number of abstract acitons

        Returns:
            int: the total number of abstract actions of doudizhu
        '''
        return 27472

    def get_player_id(self):
        ''' Return current player's id

        Returns:
            int: current player's id
        '''
        return self.round.current_player

    def get_num_players(self):
        ''' Return the number of players in doudizhu

        Returns:
            int: the number of players in doudizhu
        '''
        return self.num_players

    def is_over(self):
        ''' Judge whether a game is over

        Returns:
            Bool: True(over) / False(not over)
        '''
        if self.winner_id is None:
            return False
        return True

    def _get_others_current_hand(self, player):
        player_up = self.players[(player.player_id+1) % len(self.players)]
        player_down = self.players[(player.player_id-1) % len(self.players)]
        others_hand = merge(player_up.current_hand, player_down.current_hand, key=functools.cmp_to_key(doudizhu_sort_card))
        return cards2str(others_hand)



'''
    File name: tests/envs/test_gin_rummy_env.py
    Author: William Hale
    Date created: 4/20/2020
'''

import unittest
import numpy as np

import rlcard
from rlcard.agents.random_agent import RandomAgent
from .determism_util import is_deterministic


class TestGinRummyEnv(unittest.TestCase):

    def test_reset_and_extract_state(self):
        env = rlcard.make('gin-rummy')
        state, _ = env.reset()
        self.assertEqual(state['obs'].size, 5 * 52)

    def test_is_deterministic(self):
        self.assertTrue(is_deterministic('gin-rummy'))

    def test_get_legal_actions(self):
        env = rlcard.make('gin-rummy')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_players)])
        env.reset()
        legal_actions = env._get_legal_actions()
        for legal_action in legal_actions:
            self.assertLessEqual(legal_action, env.num_actions-1)

    def test_step(self):
        env = rlcard.make('gin-rummy')
        state, _ = env.reset()
        action = np.random.choice(list(state['legal_actions'].keys()))
        _, player_id = env.step(action)
        current_player_id = env.game.round.get_current_player().player_id
        self.assertEqual(player_id, current_player_id)

    def test_run(self):
        env = rlcard.make('gin-rummy')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_players)])
        trajectories, payoffs = env.run(is_training=False)
        self.assertEqual(len(trajectories), 2)
        for payoff in payoffs:
            self.assertLessEqual(-1, payoff)
            self.assertLessEqual(payoff, 1)
        trajectories, payoffs = env.run(is_training=True)
        for payoff in payoffs:
            self.assertLessEqual(-1, payoff)
            self.assertLessEqual(payoff, 1)


if __name__ == '__main__':
    unittest.main()



'''
    Project: Gui Gin Rummy
    File name: status_messaging.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from . import configurations
from . import info_messaging
from . import utils

from .configurations import DECLARE_DEAD_HAND_ACTION_ID
from rlcard.games.gin_rummy.game import GinRummyGame


#
#   Show prolog messages
#

def show_prolog_message(player_id: int, legal_actions, game_canvas: 'GameCanvas'):
    game_canvas_moves = game_canvas.getter.get_game_canvas_moves()
    game_canvas_moves_count = len(game_canvas_moves)
    if 0 < game_canvas_moves_count < 3 and player_id == 1:
        info_messaging.show_activate_menus_message(game_canvas=game_canvas)
    if game_canvas.query.can_declare_dead_hand(player_id=player_id):
        _show_get_card_message(player_id=player_id, game_canvas=game_canvas)
    elif game_canvas.query.can_draw_from_stock_pile(player_id=player_id):
        info_messaging.show_pick_up_discard_message(player_id=player_id, game_canvas=game_canvas)
        _show_get_card_message(player_id=player_id, game_canvas=game_canvas)
    elif game_canvas.query.can_gin(player_id=player_id):
        show_put_card_message(player_id=player_id, game_canvas=game_canvas)
    elif game_canvas.query.can_discard_card(player_id=player_id):
        show_put_card_message(player_id=player_id, game_canvas=game_canvas)
    elif game_canvas.query.is_scoring(legal_actions=legal_actions):
        _show_scoring_message(game_canvas=game_canvas)


def show_put_card_message(player_id: int, game_canvas: 'GameCanvas'):
    is_human_player = game_canvas.query.is_human(player_id=player_id)
    player_name = utils.player_name(player_id=player_id)
    if game_canvas.query.is_dead_hand_button_visible():
        game_canvas.dead_hand_button.place_forget()  # hide button
    can_gin = game_canvas.query.can_gin(player_id=player_id)
    can_knock = game_canvas.query.can_knock(player_id=player_id)
    _show_going_out_button(can_gin=can_gin, can_knock=can_knock, player_id=player_id, game_canvas=game_canvas)
    info_messaging.show_arrange_cards_message(player_id=player_id, game_canvas=game_canvas)
    show_status_messages = configurations.SHOW_STATUS_MESSAGES
    if not show_status_messages == "none":
        is_verbose = show_status_messages == "verbose"
        if not is_human_player:
            message = "{} is discarding.".format(player_name)
        else:
            if is_verbose:
                if can_gin:
                    message = "Put card: tap the gin button to go out."
                elif can_knock:
                    prefix_message = "Put card: tap a card in your hand"
                    suffix_message = "tap the discard pile to discard it or tap the knock button to go out."
                    message = "{}; then {}".format(prefix_message, suffix_message)
                else:
                    prefix_message = "Put card: tap a card in your hand"
                    suffix_message = "tap the discard pile to discard it."
                    message = "{}; then {}".format(prefix_message, suffix_message)
            else:
                message = "Put card."
        game_canvas.info_label.configure(text=message)


#
#   Show epilog messages
#

def show_epilog_message_on_declare_dead_hand(game_canvas: 'GameCanvas'):
    game_canvas.info_label.configure(text="")


def show_game_over_message(game: GinRummyGame, game_canvas: 'GameCanvas'):
    move = game.round.move_sheet[-3]
    if move.action.action_id == DECLARE_DEAD_HAND_ACTION_ID:
        dead_hand_declarer = move.player
        if dead_hand_declarer.player_id == 1:
            prefix_message = "You declared the hand dead."
        else:
            prefix_message = "Your opponent declared the hand dead."
    else:
        prefix_message = "The game is over."
    message = "{} Tap the top card of the discard pile to start a new game.".format(prefix_message)
    game_canvas.info_label.configure(text=message)
    info_messaging.show_hide_tips_message(game_canvas=game_canvas)


#
#   private methods
#

def _show_get_card_message(player_id: int, game_canvas: 'GameCanvas'):
    is_human_player = game_canvas.query.is_human(player_id=player_id)
    player_name = utils.player_name(player_id=player_id)
    can_declare_dead_hand = game_canvas.query.can_declare_dead_hand(player_id=player_id)
    if player_id == 1 and can_declare_dead_hand:
        game_canvas.dead_hand_button.place(game_canvas.dead_hand_button_place)  # show button
    show_status_messages = configurations.SHOW_STATUS_MESSAGES
    if not show_status_messages == "none":
        is_verbose = show_status_messages == "verbose"
        if can_declare_dead_hand:
            if game_canvas.current_player_id == 0:
                message = "Your opponent can declare the hand is dead."
            elif not is_human_player:
                message = "You can declare the hand is dead."
            else:
                if is_verbose:
                    message = ("Tap the dead hand button to declare hand is dead. "
                               "Or tap the discard pile to pickup a discard; "
                               "then tap a card in your hand to place it.")
                else:
                    message = "You can declare the hand is dead."
        else:
            if not is_human_player:
                message = "{} is drawing a card.".format(player_name)
            else:
                if is_verbose:
                    message = ("Get card: tap the stockpile to draw a card "
                               "or tap the discard pile to pickup a discard; "
                               "then tap a card in your hand to place it.")
                else:
                    message = "Get card."
        game_canvas.info_label.configure(text=message)


def _show_going_out_button(can_gin: bool, can_knock: bool, player_id: int, game_canvas: 'GameCanvas'):
    if can_gin and player_id == 1:  # note: show gin button only for south player
        game_canvas.going_out_button.configure(text="Gin")
        game_canvas.going_out_button.place(game_canvas.going_out_button_place)  # show button
    elif can_knock and player_id == 1:  # note: show knock button only for south player
        game_canvas.going_out_button.configure(text="Knock")
        game_canvas.going_out_button.place(game_canvas.going_out_button_place)  # show button
    else:
        game_canvas.going_out_button.place_forget()  # hide button


def _show_scoring_message(game_canvas: 'GameCanvas'):
    game_canvas.info_label.configure(text="")



import json
import os
import numpy as np
from collections import OrderedDict

import rlcard
from rlcard.envs import Env
from rlcard.games.limitholdem import Game

DEFAULT_GAME_CONFIG = {
        'game_num_players': 2,
        }

class LimitholdemEnv(Env):
    ''' Limitholdem Environment
    '''

    def __init__(self, config):
        ''' Initialize the Limitholdem environment
        '''
        self.name = 'limit-holdem'
        self.default_game_config = DEFAULT_GAME_CONFIG
        self.game = Game()
        super().__init__(config)
        self.actions = ['call', 'raise', 'fold', 'check']
        self.state_shape = [[72] for _ in range(self.num_players)]
        self.action_shape = [None for _ in range(self.num_players)]

        with open(os.path.join(rlcard.__path__[0], 'games/limitholdem/card2index.json'), 'r') as file:
            self.card2index = json.load(file)

    def _get_legal_actions(self):
        ''' Get all leagal actions

        Returns:
            encoded_action_list (list): return encoded legal action list (from str to int)
        '''
        return self.game.get_legal_actions()

    def _extract_state(self, state):
        ''' Extract the state representation from state dictionary for agent

        Note: Currently the use the hand cards and the public cards. TODO: encode the states

        Args:
            state (dict): Original state from the game

        Returns:
            observation (list): combine the player's score and dealer's observable score for observation
        '''
        extracted_state = {}

        legal_actions = OrderedDict({self.actions.index(a): None for a in state['legal_actions']})
        extracted_state['legal_actions'] = legal_actions

        public_cards = state['public_cards']
        hand = state['hand']
        raise_nums = state['raise_nums']
        cards = public_cards + hand
        idx = [self.card2index[card] for card in cards]
        obs = np.zeros(72)
        obs[idx] = 1
        for i, num in enumerate(raise_nums):
            obs[52 + i * 5 + num] = 1
        extracted_state['obs'] = obs

        extracted_state['raw_obs'] = state
        extracted_state['raw_legal_actions'] = [a for a in state['legal_actions']]
        extracted_state['action_record'] = self.action_recorder

        return extracted_state

    def get_payoffs(self):
        ''' Get the payoff of a game

        Returns:
           payoffs (list): list of payoffs
        '''
        return self.game.get_payoffs()

    def _decode_action(self, action_id):
        ''' Decode the action for applying to the game

        Args:
            action id (int): action id

        Returns:
            action (str): action for the game
        '''
        legal_actions = self.game.get_legal_actions()
        if self.actions[action_id] not in legal_actions:
            if 'check' in legal_actions:
                return 'check'
            else:
                return 'fold'
        return self.actions[action_id]

    def get_perfect_information(self):
        ''' Get the perfect information of the current state

        Returns:
            (dict): A dictionary of all the perfect information of the current state
        '''
        state = {}
        state['chips'] = [self.game.players[i].in_chips for i in range(self.num_players)]
        state['public_card'] = [c.get_index() for c in self.game.public_cards] if self.game.public_cards else None
        state['hand_cards'] = [[c.get_index() for c in self.game.players[i].hand] for i in range(self.num_players)]
        state['current_player'] = self.game.game_pointer
        state['legal_actions'] = self.game.get_legal_actions()
        return state



'''
    Project: Gui Gin Rummy
    File name: utils_extra.py
    Author: William Hale
    Date created: 3/14/2020
'''

from PIL import Image, ImageDraw, ImageFilter


def rounded_rectangle(self: ImageDraw, xy, corner_radius, fill=None, outline=None):  # FIXME: not used
    upper_left_point = xy[0]
    bottom_right_point = xy[1]
    self.rectangle(
        [
            (upper_left_point[0], upper_left_point[1] + corner_radius),
            (bottom_right_point[0], bottom_right_point[1] - corner_radius)
        ],
        fill=fill,
        outline=outline
    )
    self.rectangle(
        [
            (upper_left_point[0] + corner_radius, upper_left_point[1]),
            (bottom_right_point[0] - corner_radius, bottom_right_point[1])
        ],
        fill=fill,
        outline=outline
    )
    self.pieslice(
        [upper_left_point, (upper_left_point[0] + corner_radius * 2, upper_left_point[1] + corner_radius * 2)],
        180,
        270,
        fill=fill,
        outline=outline
        )
    self.pieslice(
        [(bottom_right_point[0] - corner_radius * 2, bottom_right_point[1] - corner_radius * 2), bottom_right_point],
        0,
        90,
        fill=fill,
        outline=outline
        )
    self.pieslice([(upper_left_point[0], bottom_right_point[1] - corner_radius * 2),
                   (upper_left_point[0] + corner_radius * 2, bottom_right_point[1])],
                  90,
                  180,
                  fill=fill,
                  outline=outline
                  )
    self.pieslice([(bottom_right_point[0] - corner_radius * 2, upper_left_point[1]),
                   (bottom_right_point[0], upper_left_point[1] + corner_radius * 2)],
                  270,
                  360,
                  fill=fill,
                  outline=outline
                  )


ImageDraw.rounded_rectangle = rounded_rectangle  # FIXME: not used


def mask_rounded_rectangle_transparent(pil_img, corner_radius=8):  # FIXME: not used
    blur_radius = 0  # FIXME: what is this for ??? wch
    mask = Image.new("L", pil_img.size, 0)
    draw = ImageDraw.Draw(mask)
    rounded_rectangle(draw, xy=((0, 0), (pil_img.size[0], pil_img.size[1])), corner_radius=corner_radius, fill=255)

    mask = mask.filter(ImageFilter.GaussianBlur(blur_radius))
    result = pil_img.copy()
    result.putalpha(mask)
    return result



from rlcard.utils import *

class Env(object):
    '''
    The base Env class. For all the environments in RLCard,
    we should base on this class and implement as many functions
    as we can.
    '''
    def __init__(self, config):
        ''' Initialize the environment

        Args:
            config (dict): A config dictionary. All the fields are
                optional. Currently, the dictionary includes:
                'seed' (int) - A environment local random seed.
                'allow_step_back' (boolean) - True if allowing
                 step_back.
                There can be some game specific configurations, e.g., the
                number of players in the game. These fields should start with
                'game_', e.g., 'game_num_players' which specify the number of
                players in the game. Since these configurations may be game-specific,
                The default settings should be put in the Env class. For example,
                the default game configurations for Blackjack should be in
                'rlcard/envs/blackjack.py'
                TODO: Support more game configurations in the future.
        '''
        self.allow_step_back = self.game.allow_step_back = config['allow_step_back']
        self.action_recorder = []

        # Game specific configurations
        # Currently only support blackjack、limit-holdem、no-limit-holdem
        # TODO support game configurations for all the games
        supported_envs = ['blackjack', 'leduc-holdem', 'limit-holdem', 'no-limit-holdem']
        if self.name in supported_envs:
            _game_config = self.default_game_config.copy()
            for key in config:
                if key in _game_config:
                    _game_config[key] = config[key]
            self.game.configure(_game_config)

        # Get the number of players/actions in this game
        self.num_players = self.game.get_num_players()
        self.num_actions = self.game.get_num_actions()

        # A counter for the timesteps
        self.timestep = 0

        # Set random seed, default is None
        self.seed(config['seed'])


    def reset(self):
        ''' Start a new game

        Returns:
            (tuple): Tuple containing:

                (numpy.array): The begining state of the game
                (int): The begining player
        '''
        state, player_id = self.game.init_game()
        self.action_recorder = []
        return self._extract_state(state), player_id

    def step(self, action, raw_action=False):
        ''' Step forward

        Args:
            action (int): The action taken by the current player
            raw_action (boolean): True if the action is a raw action

        Returns:
            (tuple): Tuple containing:

                (dict): The next state
                (int): The ID of the next player
        '''
        if not raw_action:
            action = self._decode_action(action)

        self.timestep += 1
        # Record the action for human interface
        self.action_recorder.append((self.get_player_id(), action))
        next_state, player_id = self.game.step(action)

        return self._extract_state(next_state), player_id

    def step_back(self):
        ''' Take one step backward.

        Returns:
            (tuple): Tuple containing:

                (dict): The previous state
                (int): The ID of the previous player

        Note: Error will be raised if step back from the root node.
        '''
        if not self.allow_step_back:
            raise Exception('Step back is off. To use step_back, please set allow_step_back=True in rlcard.make')

        if not self.game.step_back():
            return False

        player_id = self.get_player_id()
        state = self.get_state(player_id)

        return state, player_id

    def set_agents(self, agents):
        '''
        Set the agents that will interact with the environment.
        This function must be called before `run`.

        Args:
            agents (list): List of Agent classes
        '''
        self.agents = agents

    def run(self, is_training=False):
        '''
        Run a complete game, either for evaluation or training RL agent.

        Args:
            is_training (boolean): True if for training purpose.

        Returns:
            (tuple) Tuple containing:

                (list): A list of trajectories generated from the environment.
                (list): A list payoffs. Each entry corresponds to one player.

        Note: The trajectories are 3-dimension list. The first dimension is for different players.
              The second dimension is for different transitions. The third dimension is for the contents of each transiton
        '''
        trajectories = [[] for _ in range(self.num_players)]
        state, player_id = self.reset()

        # Loop to play the game
        trajectories[player_id].append(state)
        while not self.is_over():
            # Agent plays
            if not is_training:
                action, _ = self.agents[player_id].eval_step(state)
            else:
                action = self.agents[player_id].step(state)

            # Environment steps
            next_state, next_player_id = self.step(action, self.agents[player_id].use_raw)
            # Save action
            trajectories[player_id].append(action)

            # Set the state and player
            state = next_state
            player_id = next_player_id

            # Save state.
            if not self.game.is_over():
                trajectories[player_id].append(state)

        # Add a final state to all the players
        for player_id in range(self.num_players):
            state = self.get_state(player_id)
            trajectories[player_id].append(state)

        # Payoffs
        payoffs = self.get_payoffs()

        return trajectories, payoffs

    def is_over(self):
        ''' Check whether the curent game is over

        Returns:
            (boolean): True if current game is over
        '''
        return self.game.is_over()

    def get_player_id(self):
        ''' Get the current player id

        Returns:
            (int): The id of the current player
        '''
        return self.game.get_player_id()


    def get_state(self, player_id):
        ''' Get the state given player id

        Args:
            player_id (int): The player id

        Returns:
            (numpy.array): The observed state of the player
        '''
        return self._extract_state(self.game.get_state(player_id))

    def get_payoffs(self):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            (list): A list of payoffs for each player.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError

    def get_perfect_information(self):
        ''' Get the perfect information of the current state

        Returns:
            (dict): A dictionary of all the perfect information of the current state
        '''
        raise NotImplementedError

    def get_action_feature(self, action):
        ''' For some environments such as DouDizhu, we can have action features

        Returns:
            (numpy.array): The action features
        '''
        # By default we use one-hot encoding
        feature = np.zeros(self.num_actions, dtype=np.int8)
        feature[action] = 1
        return feature

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        self.game.np_random = self.np_random
        return seed

    def _extract_state(self, state):
        ''' Extract useful information from state for RL. Must be implemented in the child class.

        Args:
            state (dict): The raw state

        Returns:
            (numpy.array): The extracted state
        '''
        raise NotImplementedError

    def _decode_action(self, action_id):
        ''' Decode Action id to the action in the game.

        Args:
            action_id (int): The id of the action

        Returns:
            (string): The action that will be passed to the game engine.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError

    def _get_legal_actions(self):
        ''' Get all legal actions for current state.

        Returns:
            (list): A list of legal actions' id.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError



# -*- coding: utf-8 -*-
''' Implement Doudizhu Judger class
'''
import numpy as np
import collections
from itertools import combinations
from bisect import bisect_left

from rlcard.games.doudizhu.utils import CARD_RANK_STR, CARD_RANK_STR_INDEX
from rlcard.games.doudizhu.utils import cards2str, contains_cards



class DoudizhuJudger:
    ''' Determine what cards a player can play
    '''
    @staticmethod
    def chain_indexes(indexes_list):
        ''' Find chains for solos, pairs and trios by using indexes_list

        Args:
            indexes_list: the indexes of cards those have the same count, the count could be 1, 2, or 3.

        Returns:
            list of tuples: [(start_index1, length1), (start_index1, length1), ...]

        '''
        chains = []
        prev_index = -100
        count = 0
        start = None
        for i in indexes_list:
            if (i[0] >= 12): #no chains for '2BR'
                break
            if (i[0] == prev_index + 1):
                count += 1
            else:
                if (count > 1):
                    chains.append((start, count))
                count = 1
                start = i[0]
            prev_index = i[0]
        if (count > 1):
            chains.append((start, count))
        return chains

    @classmethod
    def solo_attachments(cls, hands, chain_start, chain_length, size):
        ''' Find solo attachments for trio_chain_solo_x and four_two_solo

        Args:
            hands:
            chain_start: the index of start card of the trio_chain or trio or four
            chain_length: the size of the sequence of the chain, 1 for trio_solo or four_two_solo
            size: count of solos for the attachments

        Returns:
            list of tuples: [attachment1, attachment2, ...]
                            Each attachment has two elemnts,
                            the first one contains indexes of attached cards smaller than the index of chain_start,
                            the first one contains indexes of attached cards larger than the index of chain_start
        '''
        attachments = set()
        candidates = []
        prev_card = None
        same_card_count = 0
        for card in hands:
            #dont count those cards in the chain
            if (CARD_RANK_STR_INDEX[card] >= chain_start and CARD_RANK_STR_INDEX[card] < chain_start + chain_length):
                continue
            if (card == prev_card):
                #attachments can not have bomb
                if (same_card_count == 3):
                    continue
                #attachments can not have 3 same cards consecutive with the trio (except 3 cards of '222')
                elif (same_card_count == 2 and (CARD_RANK_STR_INDEX[card] == chain_start - 1 or CARD_RANK_STR_INDEX[card] == chain_start + chain_length) and card != '2'):
                    continue
                else:
                    same_card_count += 1
            else:
                prev_card = card
                same_card_count = 1
            candidates.append(CARD_RANK_STR_INDEX[card])
        for attachment in combinations(candidates, size):
            if (attachment[-1] == 14 and attachment[-2] == 13):
                continue
            i = bisect_left(attachment, chain_start)
            attachments.add((attachment[:i], attachment[i:]))
        return list(attachments)

    @classmethod
    def pair_attachments(cls, cards_count, chain_start, chain_length, size):
        ''' Find pair attachments for trio_chain_pair_x and four_two_pair

        Args:
            cards_count:
            chain_start: the index of start card of the trio_chain or trio or four
            chain_length: the size of the sequence of the chain, 1 for trio_pair or four_two_pair
            size: count of pairs for the attachments

        Returns:
            list of tuples: [attachment1, attachment2, ...]
                            Each attachment has two elemnts,
                            the first one contains indexes of attached cards smaller than the index of chain_start,
                            the first one contains indexes of attached cards larger than the index of chain_start
        '''
        attachments = set()
        candidates = []
        for i, _ in enumerate(cards_count):
            if (i >= chain_start and i < chain_start + chain_length):
                continue
            if (cards_count[i] == 2 or cards_count[i] == 3):
                candidates.append(i)
            elif (cards_count[i] == 4):
                candidates.append(i)
        for attachment in combinations(candidates, size):
            if (attachment[-1] == 14 and attachment[-2] == 13):
                continue
            i = bisect_left(attachment, chain_start)
            attachments.add((attachment[:i], attachment[i:]))
        return list(attachments)
        
    @staticmethod
    def playable_cards_from_hand(current_hand):
        ''' Get playable cards from hand

        Returns:
            set: set of string of playable cards
        '''
        cards_dict = collections.defaultdict(int)
        for card in current_hand:
            cards_dict[card] += 1
        cards_count = np.array([cards_dict[k] for k in CARD_RANK_STR])
        playable_cards = set()

        non_zero_indexes = np.argwhere(cards_count > 0)
        more_than_1_indexes = np.argwhere(cards_count > 1)
        more_than_2_indexes = np.argwhere(cards_count > 2)
        more_than_3_indexes = np.argwhere(cards_count > 3)
        #solo
        for i in non_zero_indexes:
            playable_cards.add(CARD_RANK_STR[i[0]])
        #pair
        for i in more_than_1_indexes:
            playable_cards.add(CARD_RANK_STR[i[0]] * 2)
        #bomb, four_two_solo, four_two_pair
        for i in more_than_3_indexes:
            cards = CARD_RANK_STR[i[0]] * 4
            playable_cards.add(cards)
            for left, right in DoudizhuJudger.solo_attachments(current_hand, i[0], 1, 2):
                pre_attached = ''
                for j in left:
                    pre_attached += CARD_RANK_STR[j]
                post_attached = ''
                for j in right:
                    post_attached += CARD_RANK_STR[j]
                playable_cards.add(pre_attached + cards + post_attached)
            for left, right in DoudizhuJudger.pair_attachments(cards_count, i[0], 1, 2):
                pre_attached = ''
                for j in left:
                    pre_attached += CARD_RANK_STR[j] * 2
                post_attached = ''
                for j in right:
                    post_attached += CARD_RANK_STR[j] * 2
                playable_cards.add(pre_attached + cards + post_attached)

        #solo_chain_5 -- #solo_chain_12
        solo_chain_indexes = DoudizhuJudger.chain_indexes(non_zero_indexes)
        for (start_index, length) in solo_chain_indexes:
            s, l = start_index, length
            while(l >= 5):
                cards = ''
                curr_index = s - 1
                curr_length = 0
                while (curr_length < l and curr_length < 12):
                    curr_index += 1
                    curr_length += 1
                    cards += CARD_RANK_STR[curr_index]
                    if (curr_length >= 5):
                        playable_cards.add(cards)
                l -= 1
                s += 1

        #pair_chain_3 -- #pair_chain_10
        pair_chain_indexes = DoudizhuJudger.chain_indexes(more_than_1_indexes)
        for (start_index, length) in pair_chain_indexes:
            s, l = start_index, length
            while(l >= 3):
                cards = ''
                curr_index = s - 1
                curr_length = 0
                while (curr_length < l and curr_length < 10):
                    curr_index += 1
                    curr_length += 1
                    cards += CARD_RANK_STR[curr_index] * 2
                    if (curr_length >= 3):
                        playable_cards.add(cards)
                l -= 1
                s += 1

        #trio, trio_solo and trio_pair
        for i in more_than_2_indexes:
            playable_cards.add(CARD_RANK_STR[i[0]] * 3)
            for j in non_zero_indexes:
                if (j < i):
                    playable_cards.add(CARD_RANK_STR[j[0]] + CARD_RANK_STR[i[0]] * 3)
                elif (j > i):
                    playable_cards.add(CARD_RANK_STR[i[0]] * 3 + CARD_RANK_STR[j[0]])
            for j in more_than_1_indexes:
                if (j < i):
                    playable_cards.add(CARD_RANK_STR[j[0]] * 2 + CARD_RANK_STR[i[0]] * 3)
                elif (j > i):
                    playable_cards.add(CARD_RANK_STR[i[0]] * 3 + CARD_RANK_STR[j[0]] * 2)

        #trio_solo, trio_pair, #trio -- trio_chain_2 -- trio_chain_6; trio_solo_chain_2 -- trio_solo_chain_5; trio_pair_chain_2 -- trio_pair_chain_4
        trio_chain_indexes = DoudizhuJudger.chain_indexes(more_than_2_indexes)
        for (start_index, length) in trio_chain_indexes:
            s, l = start_index, length
            while(l >= 2):
                cards = ''
                curr_index = s - 1
                curr_length = 0
                while (curr_length < l and curr_length < 6):
                    curr_index += 1
                    curr_length += 1
                    cards += CARD_RANK_STR[curr_index] * 3

                    #trio_chain_2 to trio_chain_6
                    if (curr_length >= 2 and curr_length <= 6):
                        playable_cards.add(cards)

                    #trio_solo_chain_2 to trio_solo_chain_5
                    if (curr_length >= 2 and curr_length <= 5):
                        for left, right in DoudizhuJudger.solo_attachments(current_hand, s, curr_length, curr_length):
                            pre_attached = ''
                            for j in left:
                                pre_attached += CARD_RANK_STR[j]
                            post_attached = ''
                            for j in right:
                                post_attached += CARD_RANK_STR[j]
                            playable_cards.add(pre_attached + cards + post_attached)

                    #trio_pair_chain2 -- trio_pair_chain_4
                    if (curr_length >= 2 and curr_length <= 4):
                        for left, right in DoudizhuJudger.pair_attachments(cards_count, s, curr_length, curr_length):
                            pre_attached = ''
                            for j in left:
                                pre_attached += CARD_RANK_STR[j] * 2
                            post_attached = ''
                            for j in right:
                                post_attached += CARD_RANK_STR[j] * 2
                            playable_cards.add(pre_attached + cards + post_attached)
                l -= 1
                s += 1
        #rocket
        if (cards_count[13] and cards_count[14]):
            playable_cards.add(CARD_RANK_STR[13] + CARD_RANK_STR[14])
        return playable_cards

    def __init__(self, players, np_random):
        ''' Initilize the Judger class for Dou Dizhu
        '''
        self.playable_cards = [set() for _ in range(3)]
        self._recorded_removed_playable_cards = [[] for _ in range(3)]
        for player in players:
            player_id = player.player_id
            current_hand = cards2str(player.current_hand)
            self.playable_cards[player_id] = self.playable_cards_from_hand(current_hand)

    def calc_playable_cards(self, player):
        ''' Recalculate all legal cards the player can play according to his
        current hand.

        Args:
            player (DoudizhuPlayer object): object of DoudizhuPlayer
            init_flag (boolean): For the first time, set it True to accelerate
              the preocess.

        Returns:
            list: list of string of playable cards
        '''
        removed_playable_cards = []

        player_id = player.player_id
        current_hand = cards2str(player.current_hand)
        missed = None
        for single in player.singles:
            if single not in current_hand:
                missed = single
                break

        playable_cards = self.playable_cards[player_id].copy()

        if missed is not None:
            position = player.singles.find(missed)
            player.singles = player.singles[position+1:]
            for cards in playable_cards:
                if missed in cards or (not contains_cards(current_hand, cards)):
                    removed_playable_cards.append(cards)
                    self.playable_cards[player_id].remove(cards)
        else:
            for cards in playable_cards:
                if not contains_cards(current_hand, cards):
                    #del self.playable_cards[player_id][cards]
                    removed_playable_cards.append(cards)
                    self.playable_cards[player_id].remove(cards)
        self._recorded_removed_playable_cards[player_id].append(removed_playable_cards)
        return self.playable_cards[player_id]

    def restore_playable_cards(self, player_id):
        ''' restore playable_cards for judger for game.step_back().

        Args:
            player_id: The id of the player whose playable_cards need to be restored
        '''
        removed_playable_cards = self._recorded_removed_playable_cards[player_id].pop()
        self.playable_cards[player_id].update(removed_playable_cards)

    def get_playable_cards(self, player):
        ''' Provide all legal cards the player can play according to his
        current hand.

        Args:
            player (DoudizhuPlayer object): object of DoudizhuPlayer
            init_flag (boolean): For the first time, set it True to accelerate
              the preocess.

        Returns:
            list: list of string of playable cards
        '''
        return self.playable_cards[player.player_id]


    @staticmethod
    def judge_game(players, player_id):
        ''' Judge whether the game is over

        Args:
            players (list): list of DoudizhuPlayer objects
            player_id (int): integer of player's id

        Returns:
            (bool): True if the game is over
        '''
        player = players[player_id]
        if not player.current_hand:
            return True
        return False

    @staticmethod
    def judge_payoffs(landlord_id, winner_id):
        payoffs = np.array([0, 0, 0])
        if winner_id == landlord_id:
            payoffs[landlord_id] = 1
        else:
            for index, _ in enumerate(payoffs):
                if index != landlord_id:
                    payoffs[index] = 1
        return payoffs



# -*- coding: utf-8 -*-
"""Implement no limit texas holdem Round class"""
from enum import Enum

from rlcard.games.limitholdem import PlayerStatus


class Action(Enum):
    FOLD = 0
    CHECK_CALL = 1
    #CALL = 2
    # RAISE_3BB = 3
    RAISE_HALF_POT = 2
    RAISE_POT = 3
    # RAISE_2POT = 5
    ALL_IN = 4
    # SMALL_BLIND = 7
    # BIG_BLIND = 8


class NolimitholdemRound:
    """Round can call functions from other classes to keep the game running"""

    def __init__(self, num_players, init_raise_amount, dealer, np_random):
        """
        Initialize the round class

        Args:
            num_players (int): The number of players
            init_raise_amount (int): The min raise amount when every round starts
        """
        self.np_random = np_random
        self.game_pointer = None
        self.num_players = num_players
        self.init_raise_amount = init_raise_amount

        self.dealer = dealer

        # Count the number without raise
        # If every player agree to not raise, the round is over
        self.not_raise_num = 0

        # Count players that are not playing anymore (folded or all-in)
        self.not_playing_num = 0

        # Raised amount for each player
        self.raised = [0 for _ in range(self.num_players)]

    def start_new_round(self, game_pointer, raised=None):
        """
        Start a new bidding round

        Args:
            game_pointer (int): The game_pointer that indicates the next player
            raised (list): Initialize the chips for each player

        Note: For the first round of the game, we need to setup the big/small blind
        """
        self.game_pointer = game_pointer
        self.not_raise_num = 0
        if raised:
            self.raised = raised
        else:
            self.raised = [0 for _ in range(self.num_players)]

    def proceed_round(self, players, action):
        """
        Call functions from other classes to keep one round running

        Args:
            players (list): The list of players that play the game
            action (str/int): An legal action taken by the player

        Returns:
            (int): The game_pointer that indicates the next player
        """
        player = players[self.game_pointer]

        if action == Action.CHECK_CALL:
            diff = max(self.raised) - self.raised[self.game_pointer]
            self.raised[self.game_pointer] = max(self.raised)
            player.bet(chips=diff)
            self.not_raise_num += 1

        elif action == Action.ALL_IN:
            all_in_quantity = player.remained_chips
            self.raised[self.game_pointer] = all_in_quantity + self.raised[self.game_pointer]
            player.bet(chips=all_in_quantity)

            self.not_raise_num = 1

        elif action == Action.RAISE_POT:
            self.raised[self.game_pointer] += self.dealer.pot
            player.bet(chips=self.dealer.pot)
            self.not_raise_num = 1

        elif action == Action.RAISE_HALF_POT:
            quantity = int(self.dealer.pot / 2)
            self.raised[self.game_pointer] += quantity
            player.bet(chips=quantity)
            self.not_raise_num = 1

        elif action == Action.FOLD:
            player.status = PlayerStatus.FOLDED

        if player.remained_chips < 0:
            raise Exception("Player in negative stake")

        if player.remained_chips == 0 and player.status != PlayerStatus.FOLDED:
            player.status = PlayerStatus.ALLIN

        self.game_pointer = (self.game_pointer + 1) % self.num_players

        if player.status == PlayerStatus.ALLIN:
            self.not_playing_num += 1
            self.not_raise_num -= 1  # Because already counted in not_playing_num
        if player.status == PlayerStatus.FOLDED:
            self.not_playing_num += 1

        # Skip the folded players
        while players[self.game_pointer].status == PlayerStatus.FOLDED:
            self.game_pointer = (self.game_pointer + 1) % self.num_players

        return self.game_pointer

    def get_nolimit_legal_actions(self, players):
        """
        Obtain the legal actions for the current player

        Args:
            players (list): The players in the game

        Returns:
           (list):  A list of legal actions
        """

        full_actions = list(Action)

        # The player can always check or call
        player = players[self.game_pointer]

        diff = max(self.raised) - self.raised[self.game_pointer]
        # If the current player has no more chips after call, we cannot raise
        if diff > 0 and diff >= player.remained_chips:
            full_actions.remove(Action.RAISE_HALF_POT)
            full_actions.remove(Action.RAISE_POT)
            full_actions.remove(Action.ALL_IN)
        # Even if we can raise, we have to check remained chips
        else:
            if self.dealer.pot > player.remained_chips:
                full_actions.remove(Action.RAISE_POT)

            if int(self.dealer.pot / 2) > player.remained_chips:
                full_actions.remove(Action.RAISE_HALF_POT)

            # Can't raise if the total raise amount is leq than the max raise amount of this round
            # If raise by pot, there is no such concern
            if Action.RAISE_HALF_POT in full_actions and \
                int(self.dealer.pot / 2) + self.raised[self.game_pointer] <= max(self.raised):
                full_actions.remove(Action.RAISE_HALF_POT)

        return full_actions

    def is_over(self):
        """
        Check whether the round is over

        Returns:
            (boolean): True if the current round is over
        """
        if self.not_raise_num + self.not_playing_num >= self.num_players:
            return True
        return False



'''
    File name: gin_rummy/settings.py
    Author: William Hale
    Date created: 2/16/2020
'''

from typing import Dict, Any

from enum import Enum


class DealerForRound(int, Enum):
    North = 0
    South = 1
    Random = 2


class Setting(str, Enum):
    dealer_for_round = "dealer_for_round"
    stockpile_dead_card_count = "stockpile_dead_card_count"
    going_out_deadwood_count = "going_out_deadwood_count"
    max_drawn_card_count = "max_drawn_card_count"
    max_move_count = "max_move_count"
    is_allowed_knock = "is_allowed_knock"
    is_allowed_gin = "is_allowed_gin"
    is_allowed_pick_up_discard = "is_allowed_pick_up_discard"
    is_allowed_to_discard_picked_up_card = "is_allowed_to_discard_picked_up_card"
    is_always_knock = "is_always_knock"
    is_south_never_knocks = "is_south_never_knocks"

    @staticmethod
    def default_setting() -> Dict['Setting', Any]:
        return {Setting.dealer_for_round: DealerForRound.Random,
                Setting.stockpile_dead_card_count: 2,
                Setting.going_out_deadwood_count: 10,  # Can specify going_out_deadwood_count before running game.
                Setting.max_drawn_card_count: 52,
                Setting.max_move_count: 200,  # prevent unlimited number of moves in a game
                Setting.is_allowed_knock: True,
                Setting.is_allowed_gin: True,
                Setting.is_allowed_pick_up_discard: True,
                Setting.is_allowed_to_discard_picked_up_card: False,
                Setting.is_always_knock: False,
                Setting.is_south_never_knocks: False
                }

    @staticmethod
    def simple_gin_rummy_setting():  # speeding up training 200213
        # North should be agent being trained.
        # North always deals.
        # South never knocks.
        # North always knocks when can.
        return {Setting.dealer_for_round: DealerForRound.North,
                Setting.stockpile_dead_card_count: 2,
                Setting.going_out_deadwood_count: 10,  # Can specify going_out_deadwood_count before running game.
                Setting.max_drawn_card_count: 52,
                Setting.max_move_count: 200,  # prevent unlimited number of moves in a game
                Setting.is_allowed_knock: True,
                Setting.is_allowed_gin: True,
                Setting.is_allowed_pick_up_discard: True,
                Setting.is_allowed_to_discard_picked_up_card: False,
                Setting.is_always_knock: True,
                Setting.is_south_never_knocks: True
                }


dealer_for_round = Setting.dealer_for_round
stockpile_dead_card_count = Setting.stockpile_dead_card_count
going_out_deadwood_count = Setting.going_out_deadwood_count
max_drawn_card_count = Setting.max_drawn_card_count
max_move_count = Setting.max_move_count
is_allowed_knock = Setting.is_allowed_knock
is_allowed_gin = Setting.is_allowed_gin
is_allowed_pick_up_discard = Setting.is_allowed_pick_up_discard
is_allowed_to_discard_picked_up_card = Setting.is_allowed_to_discard_picked_up_card
is_always_knock = Setting.is_always_knock
is_south_never_knocks = Setting.is_south_never_knocks


class Settings(object):

    def __init__(self):
        self.scorer_name = "GinRummyScorer"
        default_setting = Setting.default_setting()
        self.dealer_for_round = default_setting[Setting.dealer_for_round]
        self.stockpile_dead_card_count = default_setting[Setting.stockpile_dead_card_count]
        self.going_out_deadwood_count = default_setting[Setting.going_out_deadwood_count]
        self.max_drawn_card_count = default_setting[Setting.max_drawn_card_count]
        self.max_move_count = default_setting[Setting.max_move_count]
        self.is_allowed_knock = default_setting[Setting.is_allowed_knock]
        self.is_allowed_gin = default_setting[Setting.is_allowed_gin]
        self.is_allowed_pick_up_discard = default_setting[Setting.is_allowed_pick_up_discard]
        self.is_allowed_to_discard_picked_up_card = default_setting[Setting.is_allowed_to_discard_picked_up_card]
        self.is_always_knock = default_setting[Setting.is_always_knock]
        self.is_south_never_knocks = default_setting[Setting.is_south_never_knocks]

    def change_settings(self, config: Dict[Setting, Any]):
        corrected_config = Settings.get_config_with_invalid_settings_set_to_default_value(config=config)
        for key, value in corrected_config.items():
            if key == Setting.dealer_for_round:
                self.dealer_for_round = value
            elif key == Setting.stockpile_dead_card_count:
                self.stockpile_dead_card_count = value
            elif key == Setting.going_out_deadwood_count:
                self.going_out_deadwood_count = value
            elif key == Setting.max_drawn_card_count:
                self.max_drawn_card_count = value
            elif key == Setting.max_move_count:
                self.max_move_count = value
            elif key == Setting.is_allowed_knock:
                self.is_allowed_knock = value
            elif key == Setting.is_allowed_gin:
                self.is_allowed_gin = value
            elif key == Setting.is_allowed_pick_up_discard:
                self.is_allowed_pick_up_discard = value
            elif key == Setting.is_allowed_to_discard_picked_up_card:
                self.is_allowed_to_discard_picked_up_card = value
            elif key == Setting.is_always_knock:
                self.is_always_knock = value
            elif key == Setting.is_south_never_knocks:
                self.is_south_never_knocks = value

    def print_settings(self):
        print("========== Settings ==========")
        print("scorer_name={}".format(self.scorer_name))
        print("dealer_for_round={}".format(self.dealer_for_round))
        print("stockpile_dead_card_count={}".format(self.stockpile_dead_card_count))
        print("going_out_deadwood_count={}".format(self.going_out_deadwood_count))
        print("max_drawn_card_count={}".format(self.max_drawn_card_count))
        print("max_move_count={}".format(self.max_move_count))

        print("is_allowed_knock={}".format(self.is_allowed_knock))
        print("is_allowed_gin={}".format(self.is_allowed_gin))
        print("is_allowed_pick_up_discard={}".format(self.is_allowed_pick_up_discard))

        print("is_allowed_to_discard_picked_up_card={}".format(self.is_allowed_to_discard_picked_up_card))

        print("is_always_knock={}".format(self.is_always_knock))
        print("is_south_never_knocks={}".format(self.is_south_never_knocks))
        print("==============================")

    @staticmethod
    def get_config_with_invalid_settings_set_to_default_value(config: Dict[Setting, Any]) -> Dict[Setting, Any]:
        # Set each invalid setting to its default_value.
        result = config.copy()
        default_setting = Setting.default_setting()
        for key, value in config.items():
            if key == dealer_for_round and not isinstance(value, DealerForRound):
                result[dealer_for_round] = default_setting[dealer_for_round]
            elif key == stockpile_dead_card_count and not isinstance(value, int):
                result[stockpile_dead_card_count] = default_setting[stockpile_dead_card_count]
            elif key == going_out_deadwood_count and not isinstance(value, int):
                result[going_out_deadwood_count] = default_setting[going_out_deadwood_count]
            elif key == max_drawn_card_count and not isinstance(value, int):
                result[max_drawn_card_count] = default_setting[max_drawn_card_count]
            elif key == max_move_count and not isinstance(value, int):
                result[max_move_count] = default_setting[max_move_count]
            elif key == is_allowed_knock and not isinstance(value, bool):
                result[is_allowed_knock] = default_setting[is_allowed_knock]
            elif key == is_allowed_gin and not isinstance(value, bool):
                result[is_allowed_gin] = default_setting[is_allowed_gin]
            elif key == is_allowed_pick_up_discard and not isinstance(value, bool):
                result[is_allowed_pick_up_discard] = default_setting[is_allowed_pick_up_discard]
            elif key == is_allowed_to_discard_picked_up_card and not isinstance(value, bool):
                result[is_allowed_to_discard_picked_up_card] = default_setting[is_allowed_to_discard_picked_up_card]
            elif key == is_always_knock and not isinstance(value, bool):
                result[is_always_knock] = default_setting[is_always_knock]
            elif key == is_south_never_knocks and not isinstance(value, bool):
                result[is_south_never_knocks] = default_setting[is_south_never_knocks]
        return result




'''
    File name: models/bridge_rule_models.py
    Author: William Hale
    Date created: 11/27/2021

    Bridge rule models
'''

import numpy as np

from rlcard.games.bridge.utils.action_event import ActionEvent


class BridgeDefenderNoviceRuleAgent(object):
    '''
        Agent always passes during bidding
    '''

    def __init__(self):
        self.use_raw = False

    @staticmethod
    def step(state) -> int:
        ''' Predict the action given the current state.
            Defender Novice strategy:
                Case during make call:
                    Always choose PassAction.
                Case during play card:
                    Choose a random action.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action_id (int): the action_id predicted
        '''
        legal_action_ids = state['raw_legal_actions']
        if ActionEvent.pass_action_id in legal_action_ids:
            selected_action_id = ActionEvent.pass_action_id
        else:
            selected_action_id = np.random.choice(legal_action_ids)
        return selected_action_id

    def eval_step(self, state):
        ''' Predict the action given the current state for evaluation.
            Since the agents is not trained, this function is equivalent to step function.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action_id (int): the action_id predicted by the agent
            probabilities (list): The list of action probabilities
        '''
        probabilities = []
        return self.step(state), probabilities



import unittest
import numpy as np

import rlcard
from rlcard.agents.cfr_agent import CFRAgent

class TestNFSP(unittest.TestCase):

    def test_train(self):

        env = rlcard.make('leduc-holdem', config={'allow_step_back':True})
        agent = CFRAgent(env, model_path='experiments/cfr_model')

        for _ in range(100):
            agent.train()

        state = {'obs': np.array([1., 1., 0., 0., 0., 0.]), 'legal_actions': {0: None,2: None}, 'raw_legal_actions': ['call', 'fold']}
        action, _ = agent.eval_step(state)

        self.assertIn(action, [0, 2])

    def test_save_and_load(self):
        env = rlcard.make('leduc-holdem', config={'allow_step_back':True})
        agent = CFRAgent(env, model_path='experiments/cfr_model')

        for _ in range(100):
            agent.train()

        agent.save()

        new_agent = CFRAgent(env, model_path='experiments/cfr_model')
        new_agent.load()
        self.assertEqual(len(agent.policy), len(new_agent.policy))
        self.assertEqual(len(agent.average_policy), len(new_agent.average_policy))
        self.assertEqual(len(agent.regrets), len(new_agent.regrets))
        self.assertEqual(agent.iteration, new_agent.iteration)




'''
    File name: models/gin_rummy_rule_models.py
    Author: William Hale
    Date created: 2/12/2020

    Gin Rummy rule models
'''

from typing import TYPE_CHECKING
from collections import OrderedDict

if TYPE_CHECKING:
    from rlcard.core import Card

from typing import List

import numpy as np

import rlcard

from rlcard.models.model import Model

from rlcard.games.gin_rummy.utils.action_event import *

import rlcard.games.gin_rummy.utils.melding as melding
import rlcard.games.gin_rummy.utils.utils as utils


class GinRummyNoviceRuleAgent(object):
    '''
        Agent always discards highest deadwood value card
    '''

    def __init__(self):
        self.use_raw = False  # FIXME: should this be True ?

    @staticmethod
    def step(state):
        ''' Predict the action given the current state.
            Novice strategy:
                Case where can gin:
                    Choose one of the gin actions.
                Case where can knock:
                    Choose one of the knock actions.
                Case where can discard:
                    Gin if can. Knock if can.
                    Otherwise, put aside cards in some best meld cluster.
                    Choose one of the remaining cards with highest deadwood value.
                    Discard that card.
                Case otherwise:
                    Choose a random action.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action (int): the action predicted
        '''
        legal_actions = state['legal_actions']
        actions = legal_actions.copy()
        legal_action_events = [ActionEvent.decode_action(x) for x in legal_actions]
        gin_action_events = [x for x in legal_action_events if isinstance(x, GinAction)]
        knock_action_events = [x for x in legal_action_events if isinstance(x, KnockAction)]
        discard_action_events = [x for x in legal_action_events if isinstance(x, DiscardAction)]
        if gin_action_events:
            actions = [x.action_id for x in gin_action_events]
        elif knock_action_events:
            actions = [x.action_id for x in knock_action_events]
        elif discard_action_events:
            best_discards = GinRummyNoviceRuleAgent._get_best_discards(discard_action_events=discard_action_events,
                                                                       state=state)
            if best_discards:
                actions = [DiscardAction(card=card).action_id for card in best_discards]
        if type(actions) == OrderedDict:
            actions = list(actions.keys())
        return np.random.choice(actions)

    def eval_step(self, state):
        ''' Predict the action given the current state for evaluation.
            Since the agents is not trained, this function is equivalent to step function.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action (int): the action predicted by the agent
            probabilities (list): The list of action probabilities
        '''
        probabilities = []
        return self.step(state), probabilities

    @staticmethod
    def _get_best_discards(discard_action_events, state) -> List[Card]:
        best_discards = []  # type: List[Card]
        final_deadwood_count = 999
        env_hand = state['obs'][0]
        hand = utils.decode_cards(env_cards=env_hand)
        for discard_action_event in discard_action_events:
            discard_card = discard_action_event.card
            next_hand = [card for card in hand if card != discard_card]
            meld_clusters = melding.get_meld_clusters(hand=next_hand)
            deadwood_counts = []
            for meld_cluster in meld_clusters:
                deadwood_count = utils.get_deadwood_count(hand=next_hand, meld_cluster=meld_cluster)
                deadwood_counts.append(deadwood_count)
            best_deadwood_count = min(deadwood_counts,
                                      default=utils.get_deadwood_count(hand=next_hand, meld_cluster=[]))
            if best_deadwood_count < final_deadwood_count:
                final_deadwood_count = best_deadwood_count
                best_discards = [discard_card]
            elif best_deadwood_count == final_deadwood_count:
                best_discards.append(discard_card)
        return best_discards


class GinRummyNoviceRuleModel(Model):
    ''' Gin Rummy Rule Model
    '''

    def __init__(self):
        ''' Load pre-trained model
        '''
        super().__init__()
        env = rlcard.make('gin-rummy')
        rule_agent = GinRummyNoviceRuleAgent()
        self.rule_agents = [rule_agent for _ in range(env.num_players)]

    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.rule_agents



''' Game-related base classes
'''
class Card:
    '''
    Card stores the suit and rank of a single card

    Note:
        The suit variable in a standard card game should be one of [S, H, D, C, BJ, RJ] meaning [Spades, Hearts, Diamonds, Clubs, Black Joker, Red Joker]
        Similarly the rank variable should be one of [A, 2, 3, 4, 5, 6, 7, 8, 9, T, J, Q, K]
    '''
    suit = None
    rank = None
    valid_suit = ['S', 'H', 'D', 'C', 'BJ', 'RJ']
    valid_rank = ['A', '2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K']

    def __init__(self, suit, rank):
        ''' Initialize the suit and rank of a card

        Args:
            suit: string, suit of the card, should be one of valid_suit
            rank: string, rank of the card, should be one of valid_rank
        '''
        self.suit = suit
        self.rank = rank

    def __eq__(self, other):
        if isinstance(other, Card):
            return self.rank == other.rank and self.suit == other.suit
        else:
            # don't attempt to compare against unrelated types
            return NotImplemented

    def __hash__(self):
        suit_index = Card.valid_suit.index(self.suit)
        rank_index = Card.valid_rank.index(self.rank)
        return rank_index + 100 * suit_index

    def __str__(self):
        ''' Get string representation of a card.

        Returns:
            string: the combination of rank and suit of a card. Eg: AS, 5H, JD, 3C, ...
        '''
        return self.rank + self.suit

    def get_index(self):
        ''' Get index of a card.

        Returns:
            string: the combination of suit and rank of a card. Eg: 1S, 2H, AD, BJ, RJ...
        '''
        return self.suit+self.rank



import unittest
import os
import shutil

from rlcard.utils.logger import Logger

class TestLogger(unittest.TestCase):

    def test_log(self):
        log_dir = "experiments/newtest/test_log.txt"
        if os.path.exists(log_dir):
            shutil.rmtree(log_dir)
        with Logger(log_dir) as logger:
            logger.log("test text")
            logger.log_performance(1, 1)
            logger.log_performance(2, 2)
            logger.log_performance(3, 3)

if __name__ == '__main__':
    unittest.main()



# Copyright 2021 RLCard Team of Texas A&M University
# Copyright 2021 DouZero Team of Kwai
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import threading
import time
import timeit
import pprint
from collections import deque

import torch
from torch import multiprocessing as mp
from torch import nn

from .file_writer import FileWriter
from .model import DMCModel
from .pettingzoo_model import DMCModelPettingZoo
from .utils import (
    get_batch,
    create_buffers,
    create_optimizers,
    act,
    log,
)
from .pettingzoo_utils import (
    create_buffers_pettingzoo,
    act_pettingzoo,
)

def compute_loss(logits, targets):
    loss = ((logits - targets)**2).mean()
    return loss

def learn(
    position,
    actor_models,
    agent,
    batch,
    optimizer,
    training_device,
    max_grad_norm,
    mean_episode_return_buf,
    lock
):
    """Performs a learning (optimization) step."""
    device = "cuda:"+str(training_device) if training_device != "cpu" else "cpu"
    state = torch.flatten(batch['state'].to(device), 0, 1).float()
    action = torch.flatten(batch['action'].to(device), 0, 1).float()
    target = torch.flatten(batch['target'].to(device), 0, 1)
    episode_returns = batch['episode_return'][batch['done']]
    mean_episode_return_buf[position].append(torch.mean(episode_returns).to(device))

    with lock:
        values = agent.forward(state, action)
        loss = compute_loss(values, target)
        stats = {
            'mean_episode_return_'+str(position): torch.mean(torch.stack([_r for _r in mean_episode_return_buf[position]])).item(),
            'loss_'+str(position): loss.item(),
        }

        optimizer.zero_grad()
        loss.backward()
        nn.utils.clip_grad_norm_(agent.parameters(), max_grad_norm)
        optimizer.step()

        for actor_model in actor_models.values():
            actor_model.get_agent(position).load_state_dict(agent.state_dict())
        return stats


class DMCTrainer:    
    """
    Deep Monte-Carlo

    Args:
        env: RLCard environment
        load_model (boolean): Whether loading an existing model
        xpid (string): Experiment id (default: dmc)
        save_interval (int): Time interval (in minutes) at which to save the model
        num_actor_devices (int): The number devices used for simulation
        num_actors (int): Number of actors for each simulation device
        training_device (str): The index of the GPU used for training models, or `cpu`.
        savedir (string): Root dir where experiment data will be saved
        total_frames (int): Total environment frames to train for
        exp_epsilon (float): The prbability for exploration
        batch_size (int): Learner batch size
        unroll_length (int): The unroll length (time dimension)
        num_buffers (int): Number of shared-memory buffers
        num_threads (int): Number learner threads
        max_grad_norm (int): Max norm of gradients
        learning_rate (float): Learning rate
        alpha (float): RMSProp smoothing constant
        momentum (float): RMSProp momentum
        epsilon (float): RMSProp epsilon
    """
    def __init__(
        self,
        env,
        cuda="",
        is_pettingzoo_env=False,
        load_model=False,
        xpid='dmc',
        save_interval=30,
        num_actor_devices=1,
        num_actors=5,
        training_device="0",
        savedir='experiments/dmc_result',
        total_frames=100000000000,
        exp_epsilon=0.01,
        batch_size=32,
        unroll_length=100,
        num_buffers=50,
        num_threads=4,
        max_grad_norm=40,
        learning_rate=0.0001,
        alpha=0.99,
        momentum=0,
        epsilon=0.00001
    ):
        self.env = env

        self.plogger = FileWriter(
            xpid=xpid,
            rootdir=savedir,
        )

        self.checkpointpath = os.path.expandvars(
            os.path.expanduser('%s/%s/%s' % (savedir, xpid, 'model.tar')))

        self.T = unroll_length
        self.B = batch_size

        self.xpid = xpid
        self.load_model = load_model
        self.savedir = savedir
        self.save_interval = save_interval
        self.num_actor_devices = num_actor_devices
        self.num_actors = num_actors
        self.training_device = training_device
        self.total_frames = total_frames
        self.exp_epsilon = exp_epsilon
        self.num_buffers = num_buffers
        self.num_threads = num_threads
        self.max_grad_norm = max_grad_norm
        self.learning_rate =learning_rate
        self.alpha = alpha
        self.momentum = momentum
        self.epsilon = epsilon

        self.is_pettingzoo_env = is_pettingzoo_env
        if not self.is_pettingzoo_env:
            self.num_players = self.env.num_players
            self.action_shape = self.env.action_shape
            if self.action_shape[0] == None:  # One-hot encoding
                self.action_shape = [[self.env.num_actions] for _ in range(self.num_players)]

            def model_func(device):
                return DMCModel(
                    self.env.state_shape,
                    self.action_shape,
                    exp_epsilon=self.exp_epsilon,
                    device=str(device),
                )
        else:
            self.num_players = self.env.num_agents

            def model_func(device):
                return DMCModelPettingZoo(
                    self.env,
                    exp_epsilon=self.exp_epsilon,
                    device=device
                )
        self.model_func = model_func

        self.mean_episode_return_buf = [deque(maxlen=100) for _ in range(self.num_players)]

        if cuda == "": # Use CPU
            self.device_iterator = ['cpu']
            self.training_device = "cpu"
        else:
            self.device_iterator = range(num_actor_devices)

    def start(self):
        # Initialize actor models
        models = {}
        for device in self.device_iterator:
            model = self.model_func(device)
            model.share_memory()
            model.eval()
            models[device] = model

        # Initialize buffers
        if not self.is_pettingzoo_env:
            buffers = create_buffers(
                self.T,
                self.num_buffers,
                self.env.state_shape,
                self.action_shape,
                self.device_iterator,
            )
        else:
            buffers = create_buffers_pettingzoo(
                self.T,
                self.num_buffers,
                self.env,
                self.device_iterator,
            )

        # Initialize queues
        actor_processes = []
        ctx = mp.get_context('spawn')
        free_queue = {}
        full_queue = {}
        for device in self.device_iterator:
            _free_queue = [ctx.SimpleQueue() for _ in range(self.num_players)]
            _full_queue = [ctx.SimpleQueue() for _ in range(self.num_players)]
            free_queue[device] = _free_queue
            full_queue[device] = _full_queue

        # Learner model for training
        learner_model = self.model_func(self.training_device)

        # Create optimizers
        optimizers = create_optimizers(
            self.num_players,
            self.learning_rate,
            self.momentum,
            self.epsilon,
            self.alpha,
            learner_model,
        )

        # Stat Keys
        stat_keys = []
        for p in range(self.num_players):
            stat_keys.append('mean_episode_return_'+str(p))
            stat_keys.append('loss_'+str(p))
        frames, stats = 0, {k: 0 for k in stat_keys}

        # Load models if any
        if self.load_model and os.path.exists(self.checkpointpath):
            checkpoint_states = torch.load(
                    self.checkpointpath,
                    map_location="cuda:"+str(self.training_device) if self.training_device != "cpu" else "cpu"
            )
            for p in range(self.num_players):
                learner_model.get_agent(p).load_state_dict(checkpoint_states["model_state_dict"][p])
                optimizers[p].load_state_dict(checkpoint_states["optimizer_state_dict"][p])
                for device in self.device_iterator:
                    models[device].get_agent(p).load_state_dict(learner_model.get_agent(p).state_dict())
            stats = checkpoint_states["stats"]
            frames = checkpoint_states["frames"]
            log.info(f"Resuming preempted job, current stats:\n{stats}")


        # Starting actor processes
        for device in self.device_iterator:
            num_actors = self.num_actors
            for i in range(self.num_actors):
                actor = ctx.Process(
                    target=act_pettingzoo if self.is_pettingzoo_env else act,
                    args=(i, device, self.T, free_queue[device], full_queue[device], models[device], buffers[device], self.env))
                actor.start()
                actor_processes.append(actor)

        def batch_and_learn(i, device, position, local_lock, position_lock, lock=threading.Lock()):
            """Thread target for the learning process."""
            nonlocal frames, stats
            while frames < self.total_frames:
                batch = get_batch(
                    free_queue[device][position],
                    full_queue[device][position],
                    buffers[device][position],
                    self.B,
                    local_lock
                )
                _stats = learn(
                    position,
                    models,
                    learner_model.get_agent(position),
                    batch,
                    optimizers[position],
                    self.training_device,
                    self.max_grad_norm,
                    self.mean_episode_return_buf,
                    position_lock
                )

                with lock:
                    for k in _stats:
                        stats[k] = _stats[k]
                    to_log = dict(frames=frames)
                    to_log.update({k: stats[k] for k in stat_keys})
                    self.plogger.log(to_log)
                    frames += self.T * self.B

        for device in self.device_iterator:
            for m in range(self.num_buffers):
                for p in range(self.num_players):
                    free_queue[device][p].put(m)

        threads = []
        locks = {device: [threading.Lock() for _ in range(self.num_players)] for device in self.device_iterator}
        position_locks = [threading.Lock() for _ in range(self.num_players)]

        for device in self.device_iterator:
            for i in range(self.num_threads):
                for position in range(self.num_players):
                    thread = threading.Thread(
                        target=batch_and_learn,
                        name='batch-and-learn-%d' % i,
                        args=(
                            i,
                            device,
                            position,
                            locks[device][position],
                            position_locks[position])
                        )
                    thread.start()
                    threads.append(thread)

        def checkpoint(frames):
            log.info('Saving checkpoint to %s', self.checkpointpath)
            _agents = learner_model.get_agents()
            torch.save({
                'model_state_dict': [_agent.state_dict() for _agent in _agents],
                'optimizer_state_dict': [optimizer.state_dict() for optimizer in optimizers],
                "stats": stats,
                'frames': frames,
            }, self.checkpointpath)

            # Save the weights for evaluation purpose
            for position in range(self.num_players):
                model_weights_dir = os.path.expandvars(os.path.expanduser(
                    '%s/%s/%s' % (self.savedir, self.xpid, str(position)+'_'+str(frames)+'.pth')))
                torch.save(
                    learner_model.get_agent(position),
                    model_weights_dir
                )

        timer = timeit.default_timer
        try:
            last_checkpoint_time = timer() - self.save_interval * 60
            while frames < self.total_frames:
                start_frames = frames
                start_time = timer()
                time.sleep(5)

                if timer() - last_checkpoint_time > self.save_interval * 60:
                    checkpoint(frames)
                    last_checkpoint_time = timer()

                end_time = timer()
                fps = (frames - start_frames) / (end_time - start_time)
                log.info(
                    'After %i frames: @ %.1f fps Stats:\n%s',
                    frames,
                    fps,
                    pprint.pformat(stats),
                )
        except KeyboardInterrupt:
            return
        else:
            for thread in threads:
                thread.join()
            log.info('Learning finished after %d frames.', frames)

        checkpoint(frames)
        self.plogger.close()



import unittest
import torch
import numpy as np

from rlcard.agents.nfsp_agent import NFSPAgent

class TestNFSP(unittest.TestCase):

    def test_init(self):

        agent = NFSPAgent(num_actions=10,
                          state_shape=[10],
                          hidden_layers_sizes=[10,10],
                          q_mlp_layers=[10,10],
                          device=torch.device('cpu'))

        self.assertEqual(agent._num_actions, 10)

    def test_train(self):

        memory_init_size = 20
        num_steps = 1000

        agent = NFSPAgent(num_actions=2,
                          state_shape=[2],
                          hidden_layers_sizes=[10,10],
                          reservoir_buffer_capacity=50,
                          batch_size=4,
                          min_buffer_size_to_learn=memory_init_size,
                          q_replay_memory_size=50,
                          q_replay_memory_init_size=memory_init_size,
                          q_batch_size=4,
                          q_mlp_layers=[10,10],
                          device=torch.device('cpu'))

        predicted_action, _ = agent.eval_step({'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}, 'raw_legal_actions': ['call', 'raise']})
        self.assertGreaterEqual(predicted_action, 0)
        self.assertLessEqual(predicted_action, 1)

        for _ in range(num_steps):
            agent.sample_episode_policy()
            predicted_action = agent.step({'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}})
            self.assertGreaterEqual(predicted_action, 0)
            self.assertLessEqual(predicted_action, 1)

            ts = [{'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}}, np.random.randint(2), 0, {'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}, 'raw_legal_actions': ['call', 'raise']}, True]
            agent.feed(ts)



from rlcard.games.mahjong.dealer import MahjongDealer as Dealer
from rlcard.games.mahjong.card import MahjongCard as Card
from rlcard.games.mahjong.player import MahjongPlayer as Player
from rlcard.games.mahjong.judger import MahjongJudger as Judger
from rlcard.games.mahjong.round import MahjongRound as Round
from rlcard.games.mahjong.game import MahjongGame as Game




'''
    File name: bridge/round.py
    Author: William Hale
    Date created: 11/25/2021
'''

from typing import List

from .dealer import BridgeDealer
from .player import BridgePlayer

from .utils.action_event import CallActionEvent, PassAction, DblAction, RdblAction, BidAction, PlayCardAction
from .utils.move import BridgeMove, DealHandMove, PlayCardMove, MakeBidMove, MakePassMove, MakeDblMove, MakeRdblMove, CallMove
from .utils.tray import Tray


class BridgeRound:

    @property
    def dealer_id(self) -> int:
        return self.tray.dealer_id

    @property
    def vul(self):
        return self.tray.vul

    @property
    def board_id(self) -> int:
        return self.tray.board_id

    @property
    def round_phase(self):
        if self.is_over():
            result = 'game over'
        elif self.is_bidding_over():
            result = 'play card'
        else:
            result = 'make bid'
        return result

    def __init__(self, num_players: int, board_id: int, np_random):
        ''' Initialize the round class

            The round class maintains the following instances:
                1) dealer: the dealer of the round; dealer has trick_pile
                2) players: the players in the round; each player has his own hand_pile
                3) current_player_id: the id of the current player who has the move
                4) doubling_cube: 2 if contract is doubled; 4 if contract is redoubled; else 1
                5) play_card_count: count of PlayCardMoves
                5) move_sheet: history of the moves of the players (including the deal_hand_move)

            The round class maintains a list of moves made by the players in self.move_sheet.
            move_sheet is similar to a chess score sheet.
            I didn't want to call it a score_sheet since it is not keeping score.
            I could have called move_sheet just moves, but that might conflict with the name moves used elsewhere.
            I settled on the longer name "move_sheet" to indicate that it is the official list of moves being made.

        Args:
            num_players: int
            board_id: int
            np_random
        '''
        tray = Tray(board_id=board_id)
        dealer_id = tray.dealer_id
        self.tray = tray
        self.np_random = np_random
        self.dealer: BridgeDealer = BridgeDealer(self.np_random)
        self.players: List[BridgePlayer] = []
        for player_id in range(num_players):
            self.players.append(BridgePlayer(player_id=player_id, np_random=self.np_random))
        self.current_player_id: int = dealer_id
        self.doubling_cube: int = 1
        self.play_card_count: int = 0
        self.contract_bid_move: MakeBidMove or None = None
        self.won_trick_counts = [0, 0]  # count of won tricks by side
        self.move_sheet: List[BridgeMove] = []
        self.move_sheet.append(DealHandMove(dealer=self.players[dealer_id], shuffled_deck=self.dealer.shuffled_deck))

    def is_bidding_over(self) -> bool:
        ''' Return whether the current bidding is over
        '''
        is_bidding_over = True
        if len(self.move_sheet) < 5:
            is_bidding_over = False
        else:
            last_make_pass_moves: List[MakePassMove] = []
            for move in reversed(self.move_sheet):
                if isinstance(move, MakePassMove):
                    last_make_pass_moves.append(move)
                    if len(last_make_pass_moves) == 3:
                        break
                elif isinstance(move, CallMove):
                    is_bidding_over = False
                    break
                else:
                    break
        return is_bidding_over

    def is_over(self) -> bool:
        ''' Return whether the current game is over
        '''
        is_over = True
        if not self.is_bidding_over():
            is_over = False
        elif self.contract_bid_move:
            for player in self.players:
                if player.hand:
                    is_over = False
                    break
        return is_over

    def get_current_player(self) -> BridgePlayer or None:
        current_player_id = self.current_player_id
        return None if current_player_id is None else self.players[current_player_id]

    def get_trick_moves(self) -> List[PlayCardMove]:
        trick_moves: List[PlayCardMove] = []
        if self.is_bidding_over():
            if self.play_card_count > 0:
                trick_pile_count = self.play_card_count % 4
                if trick_pile_count == 0:
                    trick_pile_count = 4  # wch: note this
                for move in self.move_sheet[-trick_pile_count:]:
                    if isinstance(move, PlayCardMove):
                        trick_moves.append(move)
                if len(trick_moves) != trick_pile_count:
                    raise Exception(f'get_trick_moves: count of trick_moves={[str(move.card) for move in trick_moves]} does not equal {trick_pile_count}')
        return trick_moves

    def get_trump_suit(self) -> str or None:
        trump_suit = None
        if self.contract_bid_move:
            trump_suit = self.contract_bid_move.action.bid_suit
        return trump_suit

    def make_call(self, action: CallActionEvent):
        # when current_player takes CallActionEvent step, the move is recorded and executed
        current_player = self.players[self.current_player_id]
        if isinstance(action, PassAction):
            self.move_sheet.append(MakePassMove(current_player))
        elif isinstance(action, BidAction):
            self.doubling_cube = 1
            make_bid_move = MakeBidMove(current_player, action)
            self.contract_bid_move = make_bid_move
            self.move_sheet.append(make_bid_move)
        elif isinstance(action, DblAction):
            self.doubling_cube = 2
            self.move_sheet.append(MakeDblMove(current_player))
        elif isinstance(action, RdblAction):
            self.doubling_cube = 4
            self.move_sheet.append(MakeRdblMove(current_player))
        if self.is_bidding_over():
            if not self.is_over():
                self.current_player_id = self.get_left_defender().player_id
        else:
            self.current_player_id = (self.current_player_id + 1) % 4

    def play_card(self, action: PlayCardAction):
        # when current_player takes PlayCardAction step, the move is recorded and executed
        current_player = self.players[self.current_player_id]
        self.move_sheet.append(PlayCardMove(current_player, action))
        card = action.card
        current_player.remove_card_from_hand(card=card)
        self.play_card_count += 1
        # update current_player_id
        trick_moves = self.get_trick_moves()
        if len(trick_moves) == 4:
            trump_suit = self.get_trump_suit()
            winning_card = trick_moves[0].card
            trick_winner = trick_moves[0].player
            for move in trick_moves[1:]:
                trick_card = move.card
                trick_player = move.player
                if trick_card.suit == winning_card.suit:
                    if trick_card.card_id > winning_card.card_id:
                        winning_card = trick_card
                        trick_winner = trick_player
                elif trick_card.suit == trump_suit:
                    winning_card = trick_card
                    trick_winner = trick_player
            self.current_player_id = trick_winner.player_id
            self.won_trick_counts[trick_winner.player_id % 2] += 1
        else:
            self.current_player_id = (self.current_player_id + 1) % 4

    def get_declarer(self) -> BridgePlayer or None:
        declarer = None
        if self.contract_bid_move:
            trump_suit = self.contract_bid_move.action.bid_suit
            side = self.contract_bid_move.player.player_id % 2
            for move in self.move_sheet:
                if isinstance(move, MakeBidMove) and move.action.bid_suit == trump_suit and move.player.player_id % 2 == side:
                    declarer = move.player
                    break
        return declarer

    def get_dummy(self) -> BridgePlayer or None:
        dummy = None
        declarer = self.get_declarer()
        if declarer:
            dummy = self.players[(declarer.player_id + 2) % 4]
        return dummy

    def get_left_defender(self) -> BridgePlayer or None:
        left_defender = None
        declarer = self.get_declarer()
        if declarer:
            left_defender = self.players[(declarer.player_id + 1) % 4]
        return left_defender

    def get_right_defender(self) -> BridgePlayer or None:
        right_defender = None
        declarer = self.get_declarer()
        if declarer:
            right_defender = self.players[(declarer.player_id + 3) % 4]
        return right_defender

    def get_perfect_information(self):
        state = {}
        last_call_move = None
        if not self.is_bidding_over() or self.play_card_count == 0:
            last_move = self.move_sheet[-1]
            if isinstance(last_move, CallMove):
                last_call_move = last_move
        trick_moves = [None, None, None, None]
        if self.is_bidding_over():
            for trick_move in self.get_trick_moves():
                trick_moves[trick_move.player.player_id] = trick_move.card
        state['move_count'] = len(self.move_sheet)
        state['tray'] = self.tray
        state['current_player_id'] = self.current_player_id
        state['round_phase'] = self.round_phase
        state['last_call_move'] = last_call_move
        state['doubling_cube'] = self.doubling_cube
        state['contact'] = self.contract_bid_move if self.is_bidding_over() and self.contract_bid_move else None
        state['hands'] = [player.hand for player in self.players]
        state['trick_moves'] = trick_moves
        return state

    def print_scene(self):
        print(f'===== Board: {self.tray.board_id} move: {len(self.move_sheet)} player: {self.players[self.current_player_id]} phase: {self.round_phase} =====')
        print(f'dealer={self.players[self.tray.dealer_id]}')
        print(f'vul={self.vul}')
        if not self.is_bidding_over() or self.play_card_count == 0:
            last_move = self.move_sheet[-1]
            last_call_text = f'{last_move}' if isinstance(last_move, CallMove) else 'None'
            print(f'last call: {last_call_text}')
        if self.is_bidding_over() and self.contract_bid_move:
            bid_suit = self.contract_bid_move.action.bid_suit
            doubling_cube = self.doubling_cube
            if not bid_suit:
                bid_suit = 'NT'
            doubling_cube_text = "" if doubling_cube == 1 else "dbl" if doubling_cube == 2 else "rdbl"
            print(f'contract: {self.contract_bid_move.player} {self.contract_bid_move.action.bid_amount}{bid_suit} {doubling_cube_text}')
        for player in self.players:
            print(f'{player}: {[str(card) for card in player.hand]}')
        if self.is_bidding_over():
            trick_pile = ['None', 'None', 'None', 'None']
            for trick_move in self.get_trick_moves():
                trick_pile[trick_move.player.player_id] = trick_move.card
            print(f'trick_pile: {[str(card) for card in trick_pile]}')



'''
    File name: bridge/utils/tray.py
    Author: William Hale
    Date created: 11/28/2021
'''


class Tray(object):

    def __init__(self, board_id: int):
        if board_id <= 0:
            raise Exception(f'Tray: invalid board_id={board_id}')
        self.board_id = board_id

    @property
    def dealer_id(self):
        return (self.board_id - 1) % 4

    @property
    def vul(self):
        vul_none = [0, 0, 0, 0]
        vul_n_s = [1, 0, 1, 0]
        vul_e_w = [0, 1, 0, 1]
        vul_all = [1, 1, 1, 1]
        basic_vuls = [vul_none, vul_n_s, vul_e_w, vul_all]
        offset = (self.board_id - 1) // 4
        return basic_vuls[(self.board_id - 1 + offset) % 4]

    def __str__(self):
        return f'{self.board_id}: dealer_id={self.dealer_id} vul={self.vul}'



import numpy as np
from collections import OrderedDict

from rlcard.envs import Env
from rlcard.games.uno import Game
from rlcard.games.uno.utils import encode_hand, encode_target
from rlcard.games.uno.utils import ACTION_SPACE, ACTION_LIST
from rlcard.games.uno.utils import cards2list

DEFAULT_GAME_CONFIG = {
        'game_num_players': 2,
        }

class UnoEnv(Env):

    def __init__(self, config):
        self.name = 'uno'
        self.default_game_config = DEFAULT_GAME_CONFIG
        self.game = Game()
        super().__init__(config)
        self.state_shape = [[4, 4, 15] for _ in range(self.num_players)]
        self.action_shape = [None for _ in range(self.num_players)]

    def _extract_state(self, state):
        obs = np.zeros((4, 4, 15), dtype=int)
        encode_hand(obs[:3], state['hand'])
        encode_target(obs[3], state['target'])
        legal_action_id = self._get_legal_actions()
        extracted_state = {'obs': obs, 'legal_actions': legal_action_id}
        extracted_state['raw_obs'] = state
        extracted_state['raw_legal_actions'] = [a for a in state['legal_actions']]
        extracted_state['action_record'] = self.action_recorder
        return extracted_state

    def get_payoffs(self):

        return np.array(self.game.get_payoffs())

    def _decode_action(self, action_id):
        legal_ids = self._get_legal_actions()
        if action_id in legal_ids:
            return ACTION_LIST[action_id]
        # if (len(self.game.dealer.deck) + len(self.game.round.played_cards)) > 17:
        #    return ACTION_LIST[60]
        return ACTION_LIST[np.random.choice(legal_ids)]

    def _get_legal_actions(self):
        legal_actions = self.game.get_legal_actions()
        legal_ids = {ACTION_SPACE[action]: None for action in legal_actions}
        return OrderedDict(legal_ids)

    def get_perfect_information(self):
        ''' Get the perfect information of the current state

        Returns:
            (dict): A dictionary of all the perfect information of the current state
        '''
        state = {}
        state['num_players'] = self.num_players
        state['hand_cards'] = [cards2list(player.hand)
                               for player in self.game.players]
        state['played_cards'] = cards2list(self.game.round.played_cards)
        state['target'] = self.game.round.target.str
        state['current_player'] = self.game.round.current_player
        state['legal_actions'] = self.game.round.get_legal_actions(
            self.game.players, state['current_player'])
        return state



'''
    Project: Gin Rummy
    File name: gin_rummy/utils/gin_rummy_error.py
    Author: William Hale
    Date created: 4/29/2020
'''


class GinRummyError(Exception):
    pass


class GinRummyProgramError(GinRummyError):
    pass



'''
    Project: Gui Gin Rummy
    File name: game_canvas.query.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from typing import List

from rlcard.games.gin_rummy.game import GinRummyGame

from rlcard.games.gin_rummy.utils.action_event import DrawCardAction, PickUpDiscardAction, DeclareDeadHandAction
from rlcard.games.gin_rummy.utils.action_event import DiscardAction, KnockAction, GinAction
from rlcard.games.gin_rummy.utils.move import ScoreSouthMove

from . import configurations

from .configurations import SCORE_PLAYER_0_ACTION_ID, SCORE_PLAYER_1_ACTION_ID
from .player_type import PlayerType


class GameCanvasQuery(object):

    def __init__(self, game_canvas: 'GameCanvas'):
        self.game_canvas = game_canvas

    def get_game(self) -> GinRummyGame:
        return self.game_canvas.game_canvas_updater.env_thread.gin_rummy_env.game

    def is_game_over(self) -> bool:
        result = False
        game = self.get_game()
        mark = self.game_canvas.game_canvas_updater.mark
        if game.round:
            moves = game.round.move_sheet[:mark]
            if moves:
                last_move = moves[-1]
                result = isinstance(last_move, ScoreSouthMove)
        return result

    def is_human(self, player_id: int or None) -> bool:
        return False if player_id is None else self.game_canvas.player_types[player_id] is PlayerType.human_player

    def is_dead_hand_button_visible(self):
        return self.game_canvas.dead_hand_button.place_info() != {}

    def is_going_out_button_visible(self):
        return self.game_canvas.going_out_button.place_info() != {}

    def can_draw_from_stock_pile(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        draw_card_actions = [x for x in legal_actions if isinstance(x, DrawCardAction)]
        return len(draw_card_actions) > 0

    def can_draw_from_discard_pile(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        pick_up_discard_actions = [x for x in legal_actions if isinstance(x, PickUpDiscardAction)]
        return len(pick_up_discard_actions) > 0

    def can_declare_dead_hand(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        declare_dead_hand_actions = [x for x in legal_actions if isinstance(x, DeclareDeadHandAction)]
        return len(declare_dead_hand_actions) > 0

    def can_discard_card(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        discard_actions = [action for action in legal_actions if isinstance(action, DiscardAction)]
        return len(discard_actions) > 0

    def can_knock(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        knock_actions = [action for action in legal_actions if isinstance(action, KnockAction)]
        return len(knock_actions) > 0

    def can_gin(self, player_id: int) -> bool:
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=player_id)
        gin_actions = [action for action in legal_actions if isinstance(action, GinAction)]
        return len(gin_actions) > 0

    def is_top_discard_pile_item_drawn(self) -> bool:
        result = False
        top_discard_pile_item_id = self.game_canvas.getter.get_top_discard_pile_item_id()
        if top_discard_pile_item_id:
            result = configurations.DRAWN_TAG in self.game_canvas.getter.get_tags(top_discard_pile_item_id)
        return result

    def is_top_stock_pile_item_drawn(self) -> bool:
        result = False
        top_stock_pile_item_id = self.game_canvas.getter.get_top_stock_pile_item_id()
        if top_stock_pile_item_id:
            result = configurations.DRAWN_TAG in self.game_canvas.getter.get_tags(top_stock_pile_item_id)
        return result

    def is_item_id_selected(self, item_id) -> bool:
        item_tags = self.game_canvas.getter.get_tags(item_id)
        return configurations.SELECTED_TAG in item_tags

    @staticmethod
    def is_scoring(legal_actions: List[int]) -> bool:
        return SCORE_PLAYER_0_ACTION_ID in legal_actions or SCORE_PLAYER_1_ACTION_ID in legal_actions



import itertools
import unittest

from rlcard.games.limitholdem.judger import LimitHoldemJudger
from rlcard.games.limitholdem.utils import compare_hands
from rlcard.games.limitholdem.utils import Hand as Hand
import numpy as np
''' Combinations selected for testing compare_hands function
Royal straight flush ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']
Straight flush ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7']
Four of a kind ['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'] ['CT', 'ST', 'HT, 'BT', 'CK', 'C8', 'C7']
Fullhouse ['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'] ['CT', 'ST', 'HT', 'B9', 'C9', 'C8', 'C7'] ['CJ', 'SJ', 'HJ', 'B8', 'C8', 'C5', 'C7']
Flush ['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'] ['CA', 'CQ', 'CT', 'C8', 'C7', 'C4', 'C2']
Straigt ['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'] ['CK', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']
Three of a kind ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'] ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C8', 'C4'] ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'] ['ST', 'ST', 'HT', 'B9', 'C2', 'C8', 'C7']
Two_pairs ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'] ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C5', 'C7'] ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']
One_pair ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'] ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C6'] ['CT', 'ST', 'H9', 'B3', 'C2', 'C8', 'C7']
Highcards ['CJ', 'S5', 'H9', 'B4', 'C2', 'C8', 'C7'] ['CJ', 'S5', 'H9', 'B4', 'C3', 'C8', 'C7']
'''
class TestHoldemUtils(unittest.TestCase):

    def test_evaluate_hand_exception(self):

        hand = Hand(['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8'])
        with self.assertRaises(Exception):
            hand.evaluateHand()

    def test_has_high_card_false(self):

        hand = Hand(['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'S3'])
        hand.product = 20
        self.assertEqual(hand._has_high_card(), False)

    def test_compare_hands(self):

        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CQ', 'SQ', 'H9', 'B3', 'C2', 'C8', 'C6']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7'], None])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [None, ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1])
        #straight flush
        hands1 = [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'],
        ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7'],
        ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7'],
        ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7'],
        ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7']]
        winner = compare_hands(hands1)
        self.assertEqual(winner, [0, 1, 1, 1, 1])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'],
        ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1, 0])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'],
        ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7'], None])
        self.assertEqual(winner, [0, 1, 0, 0])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 1])
        #Compare between different catagories
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'], ['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7'], ['HJ', 'S5', 'C9', 'B3', 'H2', 'H8', 'H7']])
        self.assertEqual(winner, [1, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        hands2 = [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'],
        ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'],
        ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA'],
        ['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']]
        winner = compare_hands(hands2)
        self.assertEqual(winner, [0, 1, 1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CJ', 'CT', 'CQ', 'CK', 'C9', 'C8', 'CA']])
        self.assertEqual(winner, [0, 1])
        #Four of a kind
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'BJ', 'C8', 'C3', 'C4']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C3', 'C4']])
        self.assertEqual(winner, [1, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        #Fullhouse
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CT', 'ST', 'HT', 'B9', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C6']])
        self.assertEqual(winner, [1, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B5', 'C5', 'C8', 'C6']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'BT', 'CT', 'C8', 'C6']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B8', 'C8', 'C7', 'C6']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        #Flush
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        #Straight
        winner = compare_hands( [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        #Two pairs
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CT', 'ST', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CT', 'ST', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'],
        ['CT', 'ST', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7'],
        ['CJ', 'SJ', 'HT', 'BT', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C5', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C5']])
        self.assertEqual(winner, [1, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CJ', 'S5', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'BJ', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'BJ', 'CT', 'B8', 'C7']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CT', 'ST', 'HT', 'B9', 'C9', 'C8', 'C7'], ['CJ', 'SJ', 'HJ', 'B9', 'C9', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'S5', 'H9', 'B4', 'C2', 'C8', 'C7'], ['CJ', 'S6', 'H9', 'B4', 'C3', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CA', 'CQ', 'CT', 'C8', 'C6', 'C4', 'C2'], ['CA', 'CQ', 'CT', 'C8', 'C7', 'C4', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7'], ['CK', 'ST', 'HQ', 'BK', 'B9', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C7'], ['ST', 'ST', 'HT', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        #Three of a kind
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CQ', 'SQ', 'HQ', 'B9', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'BT', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'B7', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C6', 'C4']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C6', 'C4']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C7', 'C4'], ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H6', 'B6', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B9', 'C2', 'C5', 'C7'], ['CJ', 'SJ', 'H9', 'B9', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HJ', 'B9', 'C2', 'C8', 'C4'], ['CJ', 'SJ', 'HJ', 'B9', 'C3', 'C8', 'C4']])
        self.assertEqual(winner, [1, 1])
        #One pair
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CT', 'ST', 'H9', 'B3', 'C2', 'C8', 'C7']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C6']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'B3', 'C2', 'C8', 'C7'], ['CQ', 'SQ', 'H9', 'B3', 'C2', 'C8', 'C6']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'HT', 'C8', 'C7', 'B3', 'C2'], ['CJ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CJ', 'SJ', 'HT', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CJ', 'SJ', 'HT', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CJ', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2'], ['CJ', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2'], ['CJ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CJ', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2'], ['CJ', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 1])
        #high_cards
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2'], ['CQ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CQ', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CK', 'SQ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C8', 'C6', 'B3', 'C2'], ['CK', 'SQ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CK', 'SJ', 'H8', 'C7', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CK', 'SJ', 'H8', 'C7', 'C6', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C8', 'C7', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C7', 'C5', 'B3', 'C2']])
        self.assertEqual(winner, [1, 0])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C7', 'C5', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [0, 1])
        winner = compare_hands( [['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2'], ['CK', 'SJ', 'H9', 'C7', 'C6', 'B3', 'C2']])
        self.assertEqual(winner, [1, 1])
        winner = compare_hands([['C5', 'S9', 'S6', 'C2', 'CT', 'C7', 'H5'], ['S7', 'SJ', 'S6', 'C2', 'CT', 'C7', 'H5'], None, None, ['H7', 'DJ', 'S6', 'C2', 'CT', 'C7', 'H5'], None])
        self.assertEqual(winner, [0, 1, 0, 0, 1, 0])

        winner = compare_hands([['H3', 'D5', 'S6', 'H9', 'CA', 'HA', 'SA'],   # three of a kind
                                ['H2', 'H3', 'C4', 'D5', 'C6', 'S6', 'ST']])  # straight
        self.assertEqual(winner, [0, 1])

        winner = compare_hands([['H3', 'D5', 'S6', 'H9', 'CA', 'HA', 'SA'],   # three of a kind
                                ['H2', 'H3', 'C4', 'D5', 'CQ', 'SK', 'SA']])  # straight beginning with A
        self.assertEqual(winner, [0, 1])

        winner = compare_hands([['H5', 'HQ', 'C2', 'D3', 'S4', 'S5', 'HT'],   # pair
                                ['D5', 'ST', 'C2', 'D3', 'S4', 'S5', 'HA'],   # A,2,3,4,5
                                ['H6', 'HQ', 'C2', 'D3', 'S4', 'S5', 'HT'],   # 2,3,4,5,6
                                ['H4', 'HQ', 'C2', 'D3', 'S4', 'S5', 'HT'],   # pair
                                ])
        self.assertEqual(winner, [0, 0, 1, 0])

        winner = compare_hands([['D5', 'ST', 'C2', 'D3', 'S4', 'S5', 'HA'],   # A,2,3,4,5
                                ['H6', 'H7', 'CK', 'DQ', 'SJ', 'ST', 'HA'],   # T,J,Q,K,A
                                ['HA', 'HT', 'CK', 'DQ', 'SJ', 'ST', 'DT'],   # T,J,Q,K,A
                                ['H2', 'H3', 'C4', 'D2', 'CQ', 'S2', 'SA'],   # three of a kind
                                ])
        self.assertEqual(winner, [0, 1, 1, 0])

        winner = compare_hands([['H5', 'HQ', 'C2', 'D3', 'S4', 'S5', 'HT'],   # pair
                                ['D5', 'ST', 'S2', 'S3', 'S4', 'S5', 'SA'],   # A,2,3,4,5 suited
                                ['C6', 'HQ', 'C2', 'C3', 'C4', 'C5', 'HT'],   # 2,3,4,5,6 suited
                                ['H7', 'H6', 'C9', 'C3', 'C4', 'C5', 'HT'],   # 3,4,5,6,7 not suited
                                ])
        self.assertEqual(winner, [0, 0, 1, 0])
     
        winner = compare_hands([['S2', 'D8', 'H8', 'S7', 'S8', 'C8', 'D3'],  # 8-four of a kind, kicker 7
                                ['S2', 'D8', 'H8', 'S9', 'S8', 'C8', 'D3'],  # 8-four of a kind, kicker 9
                                ['H3', 'C3', 'HT', 'S3', 'ST', 'CT', 'D3'],  # 3-four of a kind, kicker T
                                ])
        self.assertEqual(winner, [0, 1, 0])

        winner = compare_hands([['CA', 'C2', 'DJ', 'CT', 'S7', 'C5', 'ST'],  # pair T, T, kicker A, J
                                ['S3', 'S4', 'DJ', 'CT', 'S7', 'C5', 'ST'],  # pair T, T, kicker J, 4
                                ['HQ', 'DA', 'DJ', 'CT', 'S7', 'C5', 'ST'],  # pair T, T, kicker A, Q
                                ['SQ', 'HA', 'DJ', 'CT', 'S7', 'C5', 'ST']   # pair T, T, kicker A, Q
                                ])
        self.assertEqual(winner, [0, 0, 1, 1])

    def test_split_pots_among_players(self):
        j = LimitHoldemJudger(np.random.RandomState(seed=7))

        # simple cases where all players bet same amount of chips
        self.assertEqual(j.split_pots_among_players([2, 2], [0, 1]), [0, 4])
        self.assertEqual(j.split_pots_among_players([2, 2], [1, 0]), [4, 0])
        self.assertEqual(j.split_pots_among_players([2, 2], [1, 1]), [2, 2])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [1, 0, 0]), [6, 0, 0])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [0, 1, 0]), [0, 6, 0])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [0, 0, 1]), [0, 0, 6])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [1, 0, 1]), [3, 0, 3])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [0, 1, 1]), [0, 3, 3])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [1, 1, 0]), [3, 3, 0])
        self.assertEqual(j.split_pots_among_players([2, 2, 2], [1, 1, 1]), [2, 2, 2])
        self.assertEqual(j.split_pots_among_players([3, 3, 3], [0, 1, 1]), [0, 4, 5])
        # for the case above 9 is not divisible by 2 so a random winner get the remainder

        # trickier cases with different amounts bet (some players are all in)
        self.assertEqual(j.split_pots_among_players([3, 2], [0, 1]), [1, 4])
        self.assertEqual(j.split_pots_among_players([3, 2], [1, 0]), [5, 0])
        self.assertEqual(j.split_pots_among_players([3, 2], [1, 1]), [3, 2])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [1, 0, 0]), [6, 2, 2])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [0, 1, 0]), [0, 10, 0])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [0, 0, 1]), [0, 0, 10])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [1, 1, 0]), [3, 7, 0])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [1, 0, 1]), [3, 0, 7])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [0, 1, 1]), [0, 5, 5])
        self.assertEqual(j.split_pots_among_players([2, 4, 4], [1, 1, 1]), [2, 4, 4])
        self.assertEqual(j.split_pots_among_players([1, 1, 2, 2, 3, 3], [0, 1, 0, 1, 0, 1]), [0, 2, 0, 4, 0, 6])

    def test_split_pots_among_players_cases_generated(self):
        def check_result(in_chips, winners, allocated):
            """check that winners have won chips (more chips in allocated than in in_chips)
               and than losers have lost chips (strictly less chips in allocated than in chips)"""
            assert sum(allocated) == sum(in_chips)
            for i in range(len(in_chips)):
                if winners[i]:
                    self.assertGreaterEqual(allocated[i], in_chips[i])
                    # can be equal for example with 2 winners and 1 loser who has bet one chip (not divisible by 2)
                    # so the winner who does not get the chip of the loser will have allocated[i] == in_chips[i]
                elif in_chips[i] > 0:
                    self.assertLess(allocated[i], in_chips[i])
                    # because there is at least one winner so a loser who bet must lose at least one chip

        randstate = np.random.RandomState(seed=7)
        j = LimitHoldemJudger(randstate)

        # test many random cases from 2 to 6 players with all winners combinations
        nb_cases = 0
        for nb_players in range(2, 7):
            for _ in range(300):
                in_chips = [randstate.randint(0, 10) for _ in range(nb_players)]
                for winners in itertools.product([0, 1], repeat=nb_players):
                    if sum(winners) == 0:
                        continue  # impossible case with no winner
                    if sum(w * v for w, v in zip(winners, in_chips)) == 0:
                        continue  # impossible case where all winners have not bet
                    allocated = j.split_pots_among_players(in_chips, winners)
                    nb_cases += 1
                    check_result(in_chips, winners, allocated)
        self.assertEqual(nb_cases, 34954)  # to check that correct number of cases have been tested


if __name__ == '__main__':
    unittest.main()
